!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!=======================================================================
! MODULE modconfig
! John Hemmings - 07 Apr 2013  
!    
! Shared MarMOT configuration data for setting up models    
!=======================================================================

  module modconfig

  implicit none

! PRECISION

  ! Precision of external reals from marmot.c 

  integer,parameter :: kind_r_ext=kind(1.0d0)       

  ! Precision of reals within MarMOT FORTRAN modules
 
  integer,parameter :: kind_r=kind(1.0d0)  

! GENERAL CONFIGURATION 

  real(kind_r_ext),parameter :: objf_nodata=-999999.
                                              ! MarMOT missing data value
  real(kind_r_ext) :: modelobjfnodata=objf_nodata
                                              ! Copy visible to marmot.c

  ! Source for atomic mass values: 
  !   Wieser & Coplen, 2011, Pure Appl. Chem. 83, 359-396

  real(kind_r_ext),parameter :: atomc=12.011  ! Atomic mass of C (g/mol) 
  real(kind_r_ext) :: modelatomc=atomc        ! Copy visible to marmot.c 

  real(kind_r),parameter :: atomn=14.007      ! Atomic mass of N (g/mol)
  real(kind_r),parameter :: atomo=15.999      ! Atomic mass of O (g/mol)
  real(kind_r),parameter :: atomsi=28.085     ! Atomic mass of Si (g/mol)
  real(kind_r),parameter :: atomca=40.078     ! Atomic mass of Ca (g/mol)
  real(kind_r),parameter :: atomfe=55.845     ! Atomic mass of Fe (g/mol)
  real(kind_r),parameter :: cnredfield=6.625  ! Redfield C:N ratio
  real(kind_r),parameter :: EpdtoW=2.52       ! Conversion from E/d to W for PAR
                                              !   waveband (Morel & Smith, 1974)
 
  integer,parameter :: vertextlen=70     ! Maximum version note/name length
  integer :: modelvertextlen=vertextlen  ! Copy visible to marmot.c 

  integer,parameter :: msgtextlen=160    ! Maximum message length
  integer :: modelmsgtextlen=msgtextlen  ! Copy visible to marmot.c 

  integer,parameter :: pnamelen=20  ! Maximum parameter name length
  integer :: modelpnamelen=pnamelen ! Copy visible to marmot.c 

  integer,parameter :: pdesclen=100 ! Max. parameter description length 
  integer :: modelpdesclen=pdesclen ! Copy visible to marmot.c 

! CONFIGURATION FOR MODEL IO 

  integer,parameter :: nk=500       ! Maximum number of model levels
  integer :: modelnk=nk             ! Copy visible to marmot.c 

! N.B. The following array sizes refer to the outer dimension of the
! external MarMOT arrays (inner dimension is 'nk' for profile arrays). 
! They define the maximum element accessible by the models and can be 
! smaller than the size in marmot.c

  integer,parameter :: nt=24        ! Number of accessible MarMOT tracers
  integer :: modelnt=nt             ! Copy visible to marmot.c
  integer,parameter :: nr=12        ! Number of accessible MarMOT ratios
  integer :: modelnr=nr             ! Copy visible to marmot.c  
  integer,parameter :: ndp=43       ! No. of accessible profile diagnostics 
  integer :: modelndp=ndp           ! Copy visible to marmot.c 
  integer,parameter :: nds=9        ! No. of accessible scalar diagnostics 
  integer :: modelnds=nds           ! Copy visible to marmot.c

! MarMOT tracer indices

  integer,parameter :: & 
    i_din=1, &                ! Dissolved inorganic nitrogen (mmolN/m^3) 
    i_nh4=2, &                ! Ammonium (mmolN/m^3) 
    i_phy=3, &                ! Phytoplankton nitrogen (mmolN/m^3) 
    i_pndia=4, &              ! Non-diatom phytop. nitrogen (mmolN/m^3) 
    i_pdiat=5, &              ! Diatom nitrogen (mmolN/m^3) 
    i_zoo=6, &                ! Zooplankton nitrogen (mmolN/m^3) 
    i_zmi=7, &                ! Microzooplankton nitrogen (mmolN/m^3) 
    i_zme=8, &                ! Mesozooplankton nitrogen (mmolN/m^3) 
    i_det=9, &                ! Detrital nitrogen (mmolN/m^3) 
    i_dic=10, &               ! Dissolved inorganic carbon (mmolC/m^3)
    i_alk=11, &               ! Alkalinity (meq/m^3)
    i_dsi=12, &               ! Silicic acid (mmolSi/m^3) 
    i_dfe=13, &               ! Dissolved iron (mmolFe/m^3)
    i_chl=14, &               ! Chlorophyll (mg/m^3)
    i_chlndia=15, &           ! Non-diatom chlorophyll (mg/m^3)
    i_chldiat=16, &           ! Diatom chlorophyll (mg/m^3)
    i_phyc=17, &              ! Phytoplankton carbon (mmolC/m^3)
    i_zooc=20, &              ! Zooplankton carbon (mmolC/m^3)
    i_pdiatsi=24, &           ! Diatom silicon (mmolSi/m^3)
    i_nitrogen=25, &          ! Nitrogen (mmolN/m^3)
    i_carbon=26, &            ! Carbon (mmolC/m^3)
    i_silicon=27, &           ! Silicon (mmolSi/m^3)
    i_iron=28, &              ! Iron (mmolFe/m^3)
    i_pon=29, &               ! PON (mmolN/m^3)
    i_poc=30                  ! POC (mmolC/m^3)

! MarMOT ratio indices

  integer,parameter :: &
    i_rcchl=1, &               ! C:Chl  
    i_rcchlndia=2, &           ! Non-diatom C:Chl  
    i_rcchldiat=3, &           ! Diatom C:Chl  
    i_rcnphy=4, &              ! Phytoplankton C:N 
    i_rcnpndia=6, &            ! Non-diatom phytoplankton C:N 
    i_rcnpdiat=5, &            ! Diatom C:N 
    i_rcnzoo=7, &              ! Zooplankton C:N
    i_rcnzmi=8, &              ! Microzooplankton C:N
    i_rcnzme=9, &              ! Mesozooplankton C:N
    i_rcndet=10, &             ! Detritus C:N
    i_rsinpdiat=11, &          ! Diatom Si:N 
    i_rfen=12                  ! Organic particles Fe:N 

! MarMOT profile diagnostic indices

  integer,parameter :: &
    i_mixed=1, &              ! Fraction of level mixed
    i_fescav=3, &             ! Fe scavenging (mmol/m^2/d)
    i_kdpar=9, &              ! Diffuse attenuation coefficient for
                              !   downwelling PAR (/m) 
    i_parfracztop=10, &        ! Fraction of surface PAR at level top ztop  
    i_parztop=11, &           ! PAR at level top ztop (E/m^2/d)
    i_parfracz=12, &          ! Fraction of surface PAR at level mid pt. z 
    i_parz=13, &              ! PAR at level mid point z (E/m^2/d)
    i_photmax=14, &           ! Max. carbon fixation rate (/d) 
    i_limdin=15, &            ! DIN limitation factor
    i_pmax=16, &              ! Light saturated carbon fixation rate for 
                              !   P-E (photosynthesis-PAR) curve (/d) 
    i_alpha=17, &             ! Initial slope of P-E curve (/(E/m^2))
    i_pz=18, &                ! Potential carbon fixation at z from P_E
                              !   curve (/d) 
    i_pk=19, &                ! Potential level mean carbon fixation 
                              !   from P_E curve (/d)  
    i_phot=20, &              ! Realized level mean carbon fixation (/d)  
    i_photmaxndia=21, &       ! Non-diatom max. carbon fixation rate (/d)
    i_limdinndia=22, &        ! Non-diatom DIN limitation factor
    i_limfendia=23, &         ! Non-diatom iron limitation factor
    i_pmaxndia=24, &          ! Non-diatom light sat. carbon fixation rate 
                              !   for P-E (photosynthesis-PAR) curve (/d)
    i_alphandia=25, &         ! Non-diatom init. slope P-E curve (/(E/m^2))
    i_pzndia=26, &            ! Non-diatom potential carbon fixation at z 
                              !   from P_E curve (/d)
    i_pkndia=27, &            ! Non-diatom potential level mean carbon 
                              !   fixation from P_E curve (/d)
    i_photndia=28, &          ! Non-diatom realized level mean carbon 
                              !   fixation (/d)
    i_photmaxdiat=29, &       ! Diatom max. carbon fixation rate (/d)
    i_limdindiat=30, &        ! Diatom DIN limitation factor
    i_limsidiat=31, &         ! Diatom silicon limitation factor
    i_limfediat=32, &         ! Diatom iron limitation factor
    i_pmaxdiat=33, &          ! Diatom light sat. carbon fixation rate for
                              !   P-E (photosynthesis-PAR) curve (/d)
    i_alphadiat=34, &         ! Diatom initial slope of P-E curve (/(E/m^2))
    i_pzdiat=35, &            ! Diatom potential carbon fixation at z from 
                              !   P_E curve (/d)
    i_pkdiat=36, &            ! Diatom potential level mean carbon fixation
                              !   from P_E curve (/d)
    i_photdiat=37, &          ! Diatom realized level mean carbon fixation (/d)
    i_wdet=38, &              ! Detrital sinking velocity (m/d, upward) 
    i_insinkn=39, &           ! N influx from sinking particles (mmol/m^3/d)
    i_insinkc=40, &           ! C influx from sinking particles (mmol/m^3/d)
    i_insinksi=41, &          ! Si influx from sinking particles (mmol/m^3/d)
    i_insinkca=42, &          ! Ca influx from sinking particles (mmol/m^3/d)
    i_insinkfe=43             ! Fe influx from sinking particles (mmol/m^3/d)

! MarMOT scalar diagnostic indices

  integer,parameter :: &
    i_rlxfrac=1, &            ! Fraction of max. relaxation rate permitted
                              !   by perturation rate limitation scheme 
    i_daylen=2, &             ! Day length (d)
    i_zeuph=3, &              ! Euphotic zone depth, 1% PAR (m) 
    i_pco2=4, &               ! Sea surface pCO2 (uatm) 
    i_co2flux=5, &            ! Air-sea CO2 flux (mmol/m^2/d, downward) 
    i_co2in=6, &              ! CO2 invasion (mmol/m^2/d) 
    i_co2out=7, &             ! CO2 evasion (mmol/m^2/d)
    i_feinsurf=8, &           ! Fe input at surface (mmol/m^2/d) 
    i_feinbed=9               ! Fe input from sea bed (mmol/m^2/d)

! CONFIGURATION FOR OCEAN COLOUR ASSIMILATION SCHEME OUTPUT

! N.B. The following array sizes refer to the outer dimension of the
! external MarMOT arrays (inner dimension is 'nk' for profile arrays). 
! They define the maximum element accessible by the models and can be 
! smaller than the size in marmot.c

  integer,parameter :: ndp_ocda=30     ! No. of accessible profile outputs
  integer :: modelndpocda=ndp_ocda     ! Copy visible to marmot.c
  integer,parameter :: nds_ocda=30     ! No. of accessible scalar outputs
  integer :: modelndsocda=nds_ocda     ! Copy visible to marmot.c

! MarMOT profile output indices for ocean colour assimilation

  integer,parameter :: &
    i_bdin=1, &
    i_bphy=2, &
    i_bzoo=3, &
    i_bdet=4, &
    i_bdic=5, &
    i_balk=6, &
    i_adin=7, &
    i_aphy=8, &
    i_azoo=9, &
    i_adet=10, &
    i_adic=11, &
    i_aalk=12, &
    i_incdin=13, &
    i_incphy=14, &
    i_inczoo=15, &
    i_incdet=16, &
    i_incdic=17, &
    i_incalk=18, &
    i_bnitrogen=19, &
    i_anitrogen=20, &
    i_incnitrogen=21, &
    i_incdin1=22, &
    i_incphy1=23, &
    i_inczoo1=24, &
    i_incdet1=25, &
    i_incdin2=26, &
    i_incphy2=27, &
    i_inczoo2=28, &
    i_incdet2=29, &
    i_ubalphy=30

! MarMOT scalar output indices for ocean colour assimilation

  integer,parameter :: &
    i_obscchl=1, &
    i_bcchl=2, &
    i_acchl=3, &
    i_inccchl=4, &
    i_obschl=5, &
    i_bchl=6, &
    i_achl=7, &
    i_incchl=8, &
    i_phy0=9, &
    i_pgrow0=10, &
    i_ploss0=11, &
    i_mld=12, &
    i_mldmax=13, &
    i_ubaldin=14, &
    i_kdinlo=15, &
    i_kdinhi=16, &
    i_kphylo=17, &
    i_kzoolo=18, &
    i_kzoohi=19, &
    i_kdetlo=20, &
    i_kmld=21, &
    i_kmldmax=22, &
    i_kmax=23, &
    i_xmin=24, &
    i_xmax=25, &
    i_pinteg=26, &
    i_exx=27, &
    i_exy=28, &
    i_exug=29, &
    i_extovr=30

   end module modconfig
