!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!=======================================================================
! MODULE hadocc
! John Hemmings - 10 Nov 2013 
!
! THIS IS A DUMMY VERSION WITH PROCESSING REMOVED FROM ALL ROUTINES
! EXCEPT codever AND getinfo
!
! Contains MarMOT interface subroutines for the HadOCC model
!
!  - codever - provide version notes to MarMOT
!  - defparm - define parameter set for MarMOT
!  - getinfo - provide parameter set specific info to MarMOT
!  - setup - set-up fixed model environment
!  - setparm - set model parameters
!  - initstate - initialize MarMOT state: derived tracer & ratios
!  - getbiotrans - output tracer-specific velocities for MarMOT advection
!  - getphotin - output data to MarMOT for photosynthesis calculation
!  - biotrend - add source-minus-sink to MarMOT transport tendencies
!  - getdiag - output model diagnostics to MarMOT
!
! MarMOT interface subroutines for ocean colour data assimilation:
!
!  - ocda_defparm - define assimilation parameter set for MarMOT
!  - ocda_setup - set up fixed environment for assimilation
!  - ocda_setparm - set assimilation scheme parameters
!  - ocda_analysis - save background state and do analysis
!  - ocda_applyinc - apply assimilation increments
!  - ocda_getdiag - output assimilation scheme report to MarMOT
! 
!=======================================================================

  module hadocc

  use modconfig, only: kind_r,kind_r_ext

  implicit none

  contains

!=======================================================================
      
  subroutine codever(version,vernote,description,descripnote)
      
  use modconfig, only: vertextlen

  implicit none

! Arguments (output)

  character(len=vertextlen+1) version      ! Code version details
  character(len=vertextlen+1) vernote      ! Code version note
  character(len=vertextlen+1) description  ! Model description document
  character(len=vertextlen+1) descripnote  ! Model description note 
      
!-----------------------------------------------------------------------

  version='NOT AVAILABLE'
  vernote='distribution subject to UKMO licence'
                              
  end subroutine

!=======================================================================

  subroutine defparm(np,pname,pdesc,pmin,pmax)
      
  use modconfig, only: pnamelen,pdesclen
      
  implicit none
      
! Arguments (output)

  integer np                   ! Number of parameters
  character(len=pnamelen+1), dimension(:) :: pname  ! Parameter names
  character(len=pdesclen+1), dimension(:) :: pdesc  ! Descriptions
  real(kind_r_ext) pmin(:)     ! Minimum values
  real(kind_r_ext) pmax(:)     ! Maximum values
      
!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine getinfo(parm,photflag,latflag,advectflag,errmsg)
        
  use modconfig, only: msgtextlen
        
  implicit none
       
! Input argument
      
  real(kind_r_ext) parm(:)      ! Model-specific parameters
        
! Output arguments

  integer photflag              ! Photosynthesis model flag
                                ! -1: external phot model prohibited
                                ! 1: external phot model required
  integer latflag               ! Latitude flag (1 if required)
  integer advectflag            ! Advection scheme flag (1 if required)
  character(len=msgtextlen+1) errmsg
                                ! Error message

!-----------------------------------------------------------------------

  errmsg='The HadOCC model is not available'

  end subroutine

!=======================================================================
      
  subroutine setup(kmax,zdz_m,lat,parm, &
                   ntinit,ntlinked,itbase,it,nrinit,ir, &
                   nphyext,reqsolav,reqmld,reqtemp,reqsst,reqsss,reqpco2atm, &
                   dsinkext)

  implicit none
      
! Input arguments

  integer kmax                    ! Number of levels
  real(kind_r_ext) zdz_m(:)       ! Depth at bottom of level (m)
  real(kind_r_ext) lat            ! Latitude (deg N) 
  real(kind_r_ext) parm(:)        ! Model-specific parameters
     
! Output arguments

  integer ntinit                  ! Max. number of tracers to be initialized
  integer ntlinked                ! Number of linked tracers
  integer itbase(:)               ! Local indices of tracers that are linked to 
  integer it(:)                   ! Indices of tracers in MarMOT array
  integer nrinit                  ! Max. number of ratios to be initialized
                                  !   external photosynthesis model
  integer ir(:)                   ! Indices of ratios in MarMOT array
  integer nphyext                 ! Number of phytoplankton types for 
  integer reqsolav                ! Flag requesting solar radiation
  integer reqmld                  ! Flag requesting mixed layer depth
  integer reqtemp                 ! Flag requesting full depth temperature
  integer reqsst                  ! Flag requesting sea-surface temperature
  integer reqsss                  ! Flag requesting sea-surface salinity
  integer reqpco2atm              ! Flag requesting atmospheric pCO2
  integer dsinkext                ! Flag indicating external handling of
                                  !   detrital sinking
!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine setparm(parm,ratio_default,linkratio_def,linkratio_min, &
                     linkratio_max)
      
  implicit none
                                 
! Input arguments
  
  real(kind_r_ext) parm(:)       ! Model-specific parameters

! Output arguments

  real(kind_r) ratio_default(:)  ! Default values for initial ratios
  real(kind_r) linkratio_def(:)  ! Default tendency ratio for linked tracers
  real(kind_r) linkratio_min(:)  ! Min. tendency ratio for linked tracers
  real(kind_r) linkratio_max(:)  ! Max. tendency ratio for linked tracers

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine initstate(tracer,ratio)

  implicit none

! Arguments (in/out)

  real(kind_r_ext) tracer(:,:)      ! MarMOT tracer array
  real(kind_r_ext) ratio(:,:)       ! MarMOT composition ratio array

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine getbiotrans(wdet)

  implicit none

! Output argument

  real(kind_r) wdet         ! Detritus velocity due to sinking (m/d, up)

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine getphotin(f_sol,f_temp,tracer,ratio,par0,pigment,alpha0, &
                       chlflag,cchl,lim_nut1,psmax_phot)

  implicit none

! Input arguments

  real(kind_r_ext) f_sol        ! Surface irradiance (W/m^2)
  real(kind_r_ext) f_temp(:)    ! Temperature (deg C) 
  real(kind_r_ext) tracer(:,:)  ! MarMOT tracer array
  real(kind_r_ext) ratio(:,:)   ! MarMOT ratio array

! Output arguments

  real(kind_r) par0             ! Surface PAR (E/m^2/d)
  real(kind_r) pigment(:)       ! Pigment conc. (mg/m^3)
  real(kind_r) alpha0           ! Initial slope of P-E curve for scalar
                                !   PAR at depth 0+ (gC/gChl/(E/m^2))
  logical chlflag(:)            ! Switch to indicate presence of chlorophyll
  real(kind_r) cchl(:)          ! C:Chl (gC/gChl)
  real(kind_r) lim_nut1(:)      ! DIN limitation factor
  real(kind_r) psmax_phot(:)    ! Phytoplankton max. growth rate (/d)

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine biotrend(tstep,tofy,yearlen,f_solav,f_mld,f_temp,f_sst,f_sss, &
                      f_pco2atm,kmphotic,phot,alpha_phot,psmax_phot, &
                      solfrac,tracer,ratio,ddt,zeuphotic)

  implicit none

! Input arguments

  real(kind_r) tstep           ! Time step for biology (d)
  real(kind_r_ext) tofy        ! Time of year (days)
  integer yearlen              ! Length of year (days)
  real(kind_r_ext) f_solav     ! Surface daily average solar irrad. (W/m2)
                               ! (Need not be daily avg. if 'l_ospectral'
                               !  is off, but a diel cycle will impact on
                               !  regenerated nutrient if 'l_ofratio' is on)
  real(kind_r_ext) f_mld       ! Mixing depth (m)
  real(kind_r_ext) f_temp(:)   ! Full depth temperature (deg C)
  real(kind_r_ext) f_sst       ! Sea surface temperature (deg C)
  real(kind_r_ext) f_sss       ! Sea surface salinity (psu)
  real(kind_r_ext) f_pco2atm   ! Atmospheric pCO2 (microatmospheres)
  integer kmphotic             ! Deepest level for photosynthesis
  real(kind_r) phot(:)         ! Level mean carbon fixation (/d)
  real(kind_r) alpha_phot(:)   ! Initial slope of P-E curve for
                               !   downwelling PAR (/(E/m^2))
  real(kind_r) psmax_phot(:)   ! Phytoplankton max. growth rate (/d)
  real(kind_r) solfrac(:)      ! Fraction of surface irrad. at level bot.
  real(kind_r_ext) tracer(:,:) ! MarMOT tracer array for bio fluxes

! In/out arguments

  real(kind_r_ext) ratio(:,:)  ! MarMOT composition ratio array
  real(kind_r_ext) ddt(:,:)    ! MarMOT tendency array (units /d)

! Output argument

  real(kind_r) zeuphotic       ! Euphotic zone depth (m)

!-----------------------------------------------------------------------

  end subroutine 

!=======================================================================

  subroutine getdiag(diagprof,diagscal)

  implicit none

! Arguments (output)

  real(kind_r_ext) diagprof(:,:)  ! MarMOT profile diagnostics array
  real(kind_r_ext) diagscal(:)    ! MarMOT scalar diagnostics array
!-----------------------------------------------------------------------

  end subroutine
  
!=======================================================================

  subroutine ocda_defparm(np,pname,pdesc,pmin,pmax)

  use modconfig, only: pnamelen,pdesclen

  implicit none

! Arguments (output)

  integer np                                       ! Number of parameters
  character(len=pnamelen+1), dimension(:) :: pname ! Parameter names
  character(len=pdesclen+1), dimension(:) :: pdesc ! Descriptions
  real(kind_r_ext) pmin(:)                         ! Minimum values
  real(kind_r_ext) pmax(:)                         ! Maximum values

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine ocda_setup(nsteplead,parm,diagflag,diagfdflag, &
                        reqmld,reqcchl,reqchl,reqpgrow,reqploss)

  implicit none
      
! Input argument

  integer nsteplead        ! Number of model steps before start of
                           ! first assimilation time step 
  real(kind_r_ext) parm(:) ! Assimilation scheme parameters
  integer diagflag         ! Flag to request scalar analysis diagnostics
  integer diagfdflag       ! Flag to request profile analysis diagnostics
      
! Output argument

  integer reqmld           ! Flag requesting mixed layer depth
  integer reqcchl          ! Flag requesting surface C:Chl
  integer reqchl           ! Flag requesting surface chlorophyll
  integer reqpgrow         ! Flag requesting phytoplankton growth rate
  integer reqploss         ! Flag requesting phytoplankton loss rate
      
!-----------------------------------------------------------------------

  end subroutine

!=======================================================================
        
  subroutine ocda_setparm(parm)
        
  implicit none
        
! Input argument
        
  real(kind_r_ext) parm(:)       ! Assimilation scheme parameters

!-----------------------------------------------------------------------
          
  end subroutine

!=======================================================================

  subroutine ocda_analysis(tstep,f_mld,f_cchl,f_chl,f_pgrow,f_ploss, &
                           tracer,ratio)

  implicit none

! Input argument

  real(kind_r) tstep               ! Model time step (d)
  real(kind_r_ext) f_mld           ! Mixing depth (m)
  real(kind_r_ext) f_cchl          ! Observed C:Chl (gC/gChl)
  real(kind_r_ext) f_chl           ! Observed chlorophyll (mg/m3)
  real(kind_r_ext) f_pgrow         ! Reference growth rate for assim.
                                   ! time step (/d) - not currently used
  real(kind_r_ext) f_ploss         ! Reference loss rate for assim.
                                   ! time step (/d) - not currently used
  real(kind_r_ext) tracer(:,:)     ! MarMOT tracer array for background

! In/out argument

  real(kind_r_ext) ratio(:,:)      ! MarMOT composition ratio array

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine ocda_applyinc(tracer)

  implicit none

! In/out argument

  real(kind_r_ext) tracer(:,:)     ! MarMOT tracer array

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  subroutine ocda_getdiag(diagprof,diagscal)

  implicit none

!  Arguments (output)

   real(kind_r_ext) diagprof(:,:)  ! MarMOT profile assim. output array
   real(kind_r_ext) diagscal(:)    ! MarMOT scalar assim. output array

!-----------------------------------------------------------------------

  end subroutine

!=======================================================================

  end module hadocc

