/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* ranwell.c - John Hemmings - 21 Nov 2013 */

/* Wrapper for the WELL1024a (1024 bit well-equidistributed long-period linear) 
   pseudo random number generator from
   F. Panneton, P. L'Ecuyer, and M. Matsumoto (2006)
   Improved Long-Period Generators Based on Linear Recurrences Modulo 2, 
   ACM Transactions on Mathematical Software, 32, 1-16. */ 

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "well1024a.h"
#include "ranwell.h"

void randominit (unsigned int seed)

/* Initializes the generator state */

{ unsigned int initstate[32];
  int i;

  if (UINT_MAX!=4294967295U)
    { fprintf(stderr,
"*** Error in 'randominit': WELLRNG1024a generator not compatible with system - requires 32-bit unsigned integers ***\n");
      abort();
    } 
  srand(seed); 
  for (i=0;i<32;i++) initstate[i]=rand();
  InitWELLRNG1024a(initstate);
}


double randomu (int printflag)

/* Returns a uniform random variate in the range 0-1 */

{ double x;

  x=WELLRNG1024a();
  if (printflag) printf("Uniform random variate: randomu = %g\n",x);
  return x;
}



