/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* tabdiff.c - John Hemmings (Aug 2005) */

/* v1.01 - Feb 2010 - corrected usage message */

#include "toolbox.h"

#define FILE1 1
#define FILE2 2
#define OUTFILE 3
#define VOFFSET 4

main (int argc, char *argv[])

{ FILE *infile1,*infile2,*outfile;
  char outfbuff[FILENAME_MAX+1];
  char *outfname;
  char *reqvarname[NFIELD_MAX];
  char varlist1[LINE_MAX+1];
  char *invarname1[NFIELD_MAX];
  char varlist2[LINE_MAX+1];
  char *invarname2[NFIELD_MAX];
  char outvarlist[LINE_MAX+1];
  char *outvarname[NFIELD_MAX];
  int reqvar1[NFIELD_MAX];
  int reqvar2[NFIELD_MAX];
  int commvar[NFIELD_MAX];
  static double x1[NFIELD_MAX];
  static char fieldlist1[LINE_MAX+1];
  static char *field1[NFIELD_MAX];
  struct datarec d1={x1,fieldlist1,field1,NULL};
  static double x2[NFIELD_MAX];
  static char fieldlist2[LINE_MAX+1];
  static char *field2[NFIELD_MAX];
  struct datarec d2={x2,fieldlist2,field2,NULL};
  static double x3[NFIELD_MAX];
  struct datarec dout={x3,NULL,NULL,NULL};
  double *value1[NFIELD_MAX];
  double *value2[NFIELD_MAX];
  int nreqvar,ninvar1,ninvar2,noutvar,found,numerr,
      status,status1,status2,lineno1,lineno2,ttb1,ttb2,ttb,i,j,len;
  double tolabs=0;
  double tolrel=0;
  double value;
  char *tol,*s;

  version("1.01",VHDR,argc,argv);

  if (argc<3)
    { fprintf(stderr,
        "Usage: %s infile1 infile2 [outfile [t=tolerance[%%]] [var1 ...]]\n",
        argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (argc<4) nreqvar=0;
  else
    { nreqvar=0;
      for (i=0;i<argc-VOFFSET;i++)
        { if (strncmp(argv[VOFFSET+i],"t=",2)==0)
            { tol=argv[VOFFSET+i]+2;
              s=tol+strlen(tol)-1;
              if (*s=='%')
                { *s='\0';
                  value=sgetnum(tol,&numerr)/100;
                  if (!numerr && value>tolrel) tolrel=value; 
                }
              else 
                { value=sgetnum(tol,&numerr);
                  if (!numerr && value>tolabs) tolabs=value; 
                }
            }
          else 
            { if (nreqvar==NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);
              reqvarname[nreqvar++]=argv[VOFFSET+i];
            }
        }
    }

  /* Get file 1 variables */

  if (strcmp(argv[FILE1],"-")==0) infile1=stdin;
  else if ((infile1=fopen(argv[FILE1],"r"))==NULL)
    exiterror(OPENINERR,argv[0],argv[FILE1],0);

  status=readvar(infile1,varlist1,invarname1,&ninvar1,&ttb1,&lineno1);
  if (status!=OK) exiterror(status,argv[0],argv[FILE1],lineno1);

  if (nreqvar>0)
    { found=indexlist(nreqvar,reqvarname,ninvar1,invarname1,reqvar1);
      if (found<nreqvar)
        { for (i=0;reqvar1[i]>=0;i++);
          exiterror(VARERR,argv[0],reqvarname[i],0);
        }
    }

  /* Get file 2 variables */

  if (strcmp(argv[FILE2],"-")==0) infile2=stdin;
  else if ((infile2=fopen(argv[FILE2],"r"))==NULL)
    exiterror(OPENINERR,argv[0],argv[FILE2],0);

  status=readvar(infile2,varlist2,invarname2,&ninvar2,&ttb2,&lineno2);
  if (status!=OK) exiterror(status,argv[0],argv[FILE2],lineno2);

  if (nreqvar>0)
    { found=indexlist(nreqvar,reqvarname,ninvar2,invarname2,reqvar2);
      if (found<nreqvar)
        { for (i=0;reqvar2[i]>=0;i++);
          exiterror(VARERR,argv[0],reqvarname[i],0);
        }
    }
  
  /* If no variables specified find common variables */

  if (nreqvar==0)
    { indexlist(ninvar1,invarname1,ninvar2,invarname2,commvar);
      for (i=0;i<ninvar1;i++)
        { if (commvar[i]>=0)
            { reqvar1[nreqvar]=i;
              reqvar2[nreqvar]=commvar[i];
              reqvarname[nreqvar]=invarname1[i];
              nreqvar++;
            }
        }
    }

  /* Output header */

  ttb=(ttb1 && ttb2);
  if (argc<4 || strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else if (strcmp(argv[OUTFILE],"-.ttb")==0) 
    { outfile=stdout;
      ttb=1;
    }
  else if ((outfile=outfopen(argv[OUTFILE],outfbuff,&outfname,&ttb))==NULL)
    exiterror(OPENOUTERR,argv[0],outfname,0);

  j=0;
  len=0;
  for (i=0;i<nreqvar;i++) len+=strlen(reqvarname[i])+2;
  if (len>LINE_MAX+1) exiterror(OUTVARERR,argv[0],NULL,0);
  s=outvarlist;
  for (i=0;i<nreqvar;i++)
    { outvarname[j++]=s;
      *(s++)='d';
      *(s++)='.';
      strcpy(s,reqvarname[i]);
      s+=strlen(s)+1;
    }

  status=writevar(outfile,ttb,nreqvar,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  /* Process data */

  for (i=0;i<nreqvar;i++)
    { value1[i]=&d1.x[reqvar1[i]];
      value2[i]=&d2.x[reqvar2[i]];
    }

  status1=readdata(infile1,ttb1,ninvar1,0,NULL,0,NULL,&d1,&lineno1);
  status2=readdata(infile2,ttb2,ninvar2,0,NULL,0,NULL,&d2,&lineno2);

  while (status1==OK && status2==OK)

    { for (i=0;i<nreqvar;i++) 
        { if (!ttb1) *value1[i]=sgetnum(d1.field[reqvar1[i]],&numerr);
          if (!ttb2) *value2[i]=sgetnum(d2.field[reqvar2[i]],&numerr);
          if (*value1[i]==NODATA || *value2[i]==NODATA) dout.x[i]=NODATA;
          else 
            { dout.x[i]=*value1[i]-*value2[i];
              if (fabs(dout.x[i])<=tolabs) dout.x[i]=0;
              if (fabs(dout.x[i]/(*value2[i]))<=tolrel) dout.x[i]=0;
            }
        }

      status=writedata(outfile,ttb,&dout,nreqvar,NULL,NULL,NULL,DEFAULT,
                       DEFAULT,1);
      if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

      status1=readdata(infile1,ttb1,ninvar1,0,NULL,0,NULL,&d1,&lineno1);
      status2=readdata(infile2,ttb2,ninvar2,0,NULL,0,NULL,&d2,&lineno2);
    }

  if (status1!=OK && status1!=EOF)
    exiterror(status1,argv[0],argv[FILE1],lineno1);
  if (status2!=OK && status2!=EOF)
    exiterror(status2,argv[0],argv[FILE2],lineno2);

  fclose(infile1);
  fclose(infile2);
  fclose(outfile);
  return EXIT_SUCCESS;
}

