/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabenum.c - John Hemmings (Oct 2000) */

/* v1.01 - Feb 2001 - Corrected bug causing duplicate numbering of 1st two 
           groups when first group keyvar(s) are zero 
   v1.02 - Mar 2010 - Used new limit NLEVEL_MAX in place of NFIELD_MAX for groups */

#include "toolbox.h"

#define INFILE 1 
#define OUTFILE 2 
#define VOFFSET 3
#define NLEVEL_MAX 20

main (int argc, char *argv[])

{ FILE *infile, *outfile;
  char outfname[FILENAME_MAX+1];
  char *fname;
  char varlist[LINE_MAX+1];
  char gnumvarlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX];
  char *byvarname[NLEVEL_MAX][NFIELD_MAX];
  char *reqvarname[NFIELD_MAX];
  char *outvarname[NFIELD_MAX];
  int byvar[NLEVEL_MAX][NFIELD_MAX];
  int reqvar[NFIELD_MAX];
  int inbyvar[NFIELD_MAX];
  int nbyvar[NFIELD_MAX];
  int byoffset[NFIELD_MAX];
  int gcount[NFIELD_MAX];
  static double x[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec d={x,fieldlist,field,NULL};
  static double x2[NFIELD_MAX];
  static char fieldlist2[LINE_MAX+1];
  static char *field2[NFIELD_MAX];
  struct datarec bycurr={x2,fieldlist2,field2,NULL};
  static double xnew[NFIELD_MAX];
  struct datarec dnew={xnew,NULL,NULL,NULL};
  int ninvar,nlevel,nreqvar,ninbyvar,noutvar,count,status,found,i,j,k,ttb,
      lineno,groupend=-1,maxchar,len;
  char *s;

  version("1.02",VHDR,argc,argv);

  if (argc<3)
    { fprintf(stderr,
"Usage: %s infile outfile [grp1keyvar1 ... [by grp2keyvar1 ... [by ...]]]\n",
              argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (argc>VOFFSET)

    { if (strcmp(argv[VOFFSET],"by")==0) byoffset[0]=VOFFSET+1;
      else byoffset[0]=VOFFSET;
      nbyvar[0]=0;
      nlevel=1;

      j=0;
      for (i=byoffset[0];i<argc;i++)
        { if (strcmp(argv[i],"by")==0)
            { if (nlevel>NLEVEL_MAX-1) exiterror(SPECIAL,argv[0],"too many group levels",0); 
              byoffset[nlevel]=i+1;
              nbyvar[nlevel]=0;
              nlevel++;
              j=0;
            }
          else
            { if (j>NFIELD_MAX-1) exiterror(OUTVARERR,argv[0],NULL,0); 
              byvarname[nlevel-1][j]=argv[i];
              nbyvar[nlevel-1]++;
              j++;
            }
        }
    }

  else nlevel=0;

  if (strcmp(argv[INFILE],"-")==0) infile=stdin;
  else if ((infile=fopen(argv[INFILE],"r"))==NULL)
    exiterror(OPENINERR,argv[0],argv[INFILE],0);

  status=readvar(infile,varlist,invarname,&ninvar,&ttb,&lineno);
  if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);

  if (nlevel!=0) 
  for (i=0;i<nlevel;i++)
    { if (indexlist(nbyvar[i],byvarname[i],ninvar,invarname,byvar[i])<nbyvar[i])
        { for (j=0;byvar[i][j]>=0;j++);
          exiterror(VARERR,argv[0],byvarname[i][j],0);
        }
    }

  ninbyvar=0;
  nreqvar=0;
  for (i=0;i<ninvar;i++)
    { j=0;
      found=0;
      while (j<nlevel && !found) 
        { k=0;
          while (k<nbyvar[j] && !(found=(byvar[j][k]==i))) k++;
          j++;
        }
      if (found) inbyvar[ninbyvar++]=i;
      else 
        { reqvar[nreqvar]=i;
          reqvarname[nreqvar]=invarname[i];
          nreqvar++;
        }
    }

  noutvar=nreqvar+1;
  for (i=0;i<nlevel;i++) noutvar+=nbyvar[i]+1;
  if (noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else
    { if (ttb) fname=ttbname(argv[OUTFILE],outfname);
      else fname=argv[OUTFILE];
      if ((outfile=fopen(fname,"w"))==NULL)
        exiterror(OPENOUTERR,argv[0],fname,0);
    }

  k=0;
  s=gnumvarlist;
  maxchar=LINE_MAX;
  for (i=nlevel-1;i>=0;i--)
    { if (maxchar<6) exiterror(OUTVARERR,argv[0],NULL,0); 
      outvarname[k++]=s;
      strcpy(s,"grpnum"); 
      s+=6;
      maxchar-=6;
      len=sputnum(s,maxchar,i+1,0,5);
      if (len<0) exiterror(OUTVARERR,argv[0],NULL,0);
      s+=len+1;
      maxchar-=len+1;
      for (j=0;j<nbyvar[i];j++) outvarname[k++]=argv[byoffset[i]+j];
    }
  outvarname[k++]="num";
  for (i=0;i<nreqvar;i++) outvarname[k++]=reqvarname[i];
  status=writevar(outfile,ttb,k,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0); 

  while ((status=readdata(infile,ttb,ninvar,ninbyvar,inbyvar,0,NULL,&d,&lineno))
                ==OK)

    { if (groupend==-1) groupend=1; 
      else groupend=0;

      for (i=nlevel-1;i>=0;i--)

        { j=i+1;
          if (groupend) gcount[i]=1;
          else if (datacomp(nbyvar[i],&d,byvar[i],&bycurr,byvar[i],NUMCOMP)!=0)
            groupend=1;

          if (groupend) dnew.x[j]=(double)gcount[i]++;

          status=
            writedata(outfile,ttb,&dnew,1,&j,NULL,NULL,DEFAULT,DEFAULT,0);
          if (status==OK)
            status=writedata(outfile,ttb,&d,nbyvar[i],byvar[i],NULL,NULL,
                             DEFAULT,DEFAULT,0);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0); 
        } 

      if (groupend) 
        { count=1;
          copydata(&bycurr,&d,ninvar,NULL);
        }
      else count++;
      dnew.x[0]=(double)count;

      status=writedata(outfile,ttb,&dnew,1,NULL,NULL,NULL,DEFAULT,DEFAULT,0);
      if (status==OK) 
        status=writedata(outfile,ttb,&d,nreqvar,reqvar,NULL,NULL,DEFAULT,
                         DEFAULT,1);
      if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0); 
    }

  if (status!=EOF) exiterror(status,argv[0],argv[INFILE],lineno);

  fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}  

