/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabfcat.c - John Hemmings (Apr 2002) */

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2
#define VOFFSET 3
#define BUFFSIZE 1000 


main (int argc, char *argv[])

{ FILE *infile, *outfile;
  char outfname[FILENAME_MAX+1];
  char *fname;
  char invarlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX],*reqvarname[NFIELD_MAX],*outvarname[NFIELD_MAX];
  int byvar[NFIELD_MAX],reqvar[NFIELD_MAX];
  static double x1[NFIELD_MAX];
  static char fieldlist1[LINE_MAX+1];
  static char *field1[NFIELD_MAX];
  struct datarec d={x1,fieldlist1,field1,NULL};
  static double x2[NFIELD_MAX];
  static char fieldlist2[LINE_MAX+1];
  static char *field2[NFIELD_MAX];
  struct datarec bycurr={x2,fieldlist2,field2,NULL};
  static char *field3[NFIELD_MAX];
  struct datarec dout={NULL,NULL,field3,NULL};
  int ninvar,noutvar,nreq,nreqvar,nbyvar,byoffset,found,status,len,seplen,
      i,j,c,ttb,lineno,first=1,fin=0;
  char *sep,*p[NFIELD_MAX];
  char buff[NFIELD_MAX][BUFFSIZE+1];
  int space[NFIELD_MAX];

  version("1.00",VHDR,argc,argv);

  if (argc<3)
    { fprintf(stderr,
        "Usage: %s infile outfile [s=separator] [var1 ...] [by keyvar1 ...]\n",
              argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  j=VOFFSET;
  found=0;
  while(j<argc && !(found=(strcmp(argv[j],"by")==0))) j++;
  if (found) 
    { byoffset=j+1;
      nbyvar=argc-byoffset;
      nreq=byoffset-VOFFSET-1;
    }
  else
    { nbyvar=0;
      nreq=argc-VOFFSET;
    }

  sep="";
  nreqvar=0;
  for (i=0;i<nreq;i++) 
    { if (strncmp(argv[VOFFSET+i],"s=",2)==0) 
        { sep=argv[VOFFSET+i]+2;
          seplen=strlen(sep);
        }
      else reqvarname[nreqvar++]=argv[VOFFSET+i];
    }
 
  if (strcmp(argv[INFILE],"-")==0) infile=stdin;
  else if ((infile=fopen(argv[INFILE],"r"))==NULL) 
    exiterror(OPENINERR,argv[0],argv[INFILE],0);

  status=readvar(infile,invarlist,invarname,&ninvar,&ttb,&lineno);
  if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);

  if (indexlist(nreqvar,reqvarname,ninvar,invarname,reqvar)<nreqvar)
    { for (i=0;reqvar[i]>=0;i++);
      exiterror(VARERR,argv[0],reqvarname[i],0);
    }
  if (indexlist(nbyvar,&argv[byoffset],ninvar,invarname,byvar)<nbyvar)
    { for (i=0;byvar[i]>=0;i++);
      exiterror(VARERR,argv[0],argv[byoffset+i],0);
    }
  
  if (nreqvar==0)
    { for (i=0;i<ninvar;i++) 
        { j=0;
          found=0;
          while (j<nbyvar && !(found=(byvar[j]==i))) j++;
          if (!found) 
            { reqvar[nreqvar]=i; 
              reqvarname[nreqvar]=invarname[i];
              nreqvar++;
            }
        }
    }
  else
    { for (i=0;i<nbyvar;i++)
        { j=0;
          found=0;
          while (j<nreqvar && !(found=(reqvar[j]==byvar[i]))) j++;
          if (found) exiterror(BADEXPR,argv[0],argv[byoffset+i],0);
        }
    }

  noutvar=nbyvar+nreqvar;
  if (noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else if ((outfile=fopen(argv[OUTFILE],"w"))==NULL)
    exiterror(OPENOUTERR,argv[0],argv[OUTFILE],0);

  j=0;
  for (i=0;i<nbyvar;i++) outvarname[j++]=argv[byoffset+i];
  for (i=0;i<nreqvar;i++) outvarname[j++]=reqvarname[i];
  status=writevar(outfile,0,noutvar,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
          
  for (i=0;i<nreqvar;i++) dout.field[i]=buff[i];

  while (!fin &&
         (status=readdata(infile,ttb,ninvar,nbyvar,byvar,nreqvar,reqvar,&d,
                          &lineno))<=OK)

    { if (status==EOF) fin=1; 

      if (!first &&
          (fin || (c=datacomp(nbyvar,&d,byvar,&bycurr,NULL,NUMCOMP))!=0))

        { status=writedata(outfile,0,&bycurr,nbyvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,0);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
          status=writedata(outfile,0,&dout,nreqvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,1);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
        }

      if (!fin)

        { if (first || c!=0)

            { for (i=0;i<nreqvar;i++) 
                { p[i]=buff[i]; 
                  space[i]=BUFFSIZE;
                }
              copydata(&bycurr,&d,nbyvar,byvar);
            }
          
          first=0;

          for (i=0;i<nreqvar;i++) 
            { len=strlen(d.field[reqvar[i]]);
              if (p[i]>buff[i]) len+=seplen;
              if (len>space[i]) 
                exiterror(OUTTRUNC,argv[0],argv[INFILE],lineno);
              if (p[i]>buff[i]) 
                { strcpy(p[i],sep);
                  len=strlen(p[i]);
                  p[i]+=len;
                  space[i]-=len;
                }
              strcpy(p[i],d.field[reqvar[i]]);
              len=strlen(p[i]);
              p[i]+=len;
              space[i]-=len;
            }
        }
    }

  if (!fin) exiterror(status,argv[0],argv[INFILE],lineno);

  fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}  
