/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* tabfilt.c - John Hemmings (Oct 2000) */

/* v1.01 - fixed bug in pointer positioning when writing trailing records
           & changed argument checks to ensure at least 1 var to process */

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2
#define WINSPEC 3
#define VOFFSET 4 

struct outspecrec {
  FILE *outfile;
  int ttb;
  int nbyvar;
  int *byvar;
  int nreqvar;
  int *reqvar;
  int nnewvar;
  struct datarec *dbuff;
} outspec;

int writeresult (struct datarec *, int [], int, double [], int, 
                 struct datarec *, struct outspecrec *);
int writeblank (struct datarec *, struct datarec *, struct outspecrec *);


main (int argc, char *argv[])

{ FILE *infile,*outfile,*wfile;
  char outfname[FILENAME_MAX+1];
  char *fname;
  char varlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX],*wvarname[NFIELD_MAX],*byvarname[NFIELD_MAX],
       *reqvarname[NFIELD_MAX],*newvarname[NFIELD_MAX],*oldvarname[NFIELD_MAX],
       *outvarname[NFIELD_MAX];
  int byvar[NFIELD_MAX],reqvar[NFIELD_MAX],oldvar[NFIELD_MAX],wtsvar;
  static double x[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec dbuff={x,fieldlist,field,NULL};
  struct datarec *d,*dnewrec,*dhead=NULL,*dwin,*dout,*djunk;
  double w[NREC_MAX],num,sum;
  int byoffset,nwvar,ninvar,nbyvar,nreqvar,nnewvar,found,i,j,status,winsize=0,
      winendfound=0,zleft=0,zright=0,winmid,nwin,nleft,nright,nhold,n,norm,
      fin=0,first=1,count=0,newgroup=1,numerr,lineno,ttb;
  char *wspec,*s;

  version("1.01",VHDR,argc,argv);

  if (argc<5)
    { fprintf(stderr,
"Usage: %s infile outfile winspec newvar1=var1 [newvar2=var2 ...]\n",argv[0]);
      fprintf(stderr,
"       [by keyvar1 ...]\n");
      fprintf(stderr,
"       where winspec is: winsize OR [=]w1,w2,w3[,w4,w5...] OR [=]file[,var]\n"
              );
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);


  /* PROCESS WINDOW SPECIFICATION */
 
  wspec=argv[WINSPEC]; 
  if (*wspec=='=') 
    { norm=0;
      wspec++;
    }
  else norm=1;     
   
  s=strchr(wspec,',');
  if (s) *(s++)='\0';

  if (strcmp(wspec,"-")==0 || isalpha(*wspec))

    { /* Read weights from file */

      if (strcmp(wspec,"-")==0) wfile=stdin;
      else if ((wfile=fopen(wspec,"r"))==NULL)
        exiterror(OPENINERR,argv[0],wspec,0);

      status=readvar(wfile,varlist,wvarname,&nwvar,&ttb,&lineno);
      if (status!=OK) exiterror(status,argv[0],wspec,lineno);
     
      if (s)
        { if (indexlist(1,&s,nwvar,wvarname,&wtsvar)<1)
            exiterror(VARERR,argv[0],s,0);
        }
      else if (nwvar!=1) exiterror(NOVAR,argv[0],NULL,0);
      else wtsvar=0;

      while ((status=readdata(wfile,ttb,nwvar,0,NULL,0,NULL,&dbuff,&lineno))
             ==OK)

        { if (winsize==NREC_MAX)
            exiterror(RECLIMIT,argv[0],wspec,lineno); 

          if (ttb) w[winsize]=dbuff.x[wtsvar]; 
          else w[winsize]=sgetnum(dbuff.field[wtsvar],&numerr); 

          if (winsize==0)
            { if (w[winsize]!=NODATA) winsize++;
            }
          else if (winendfound)
            { if (w[winsize]!=NODATA)
                exiterror(FMTERR,argv[0],wspec,lineno);
            }
          else if (w[winsize]==NODATA) winendfound=1;
          else winsize++;
        }

      if (status!=EOF) exiterror(status,argv[0],wspec,lineno);
      fclose(wfile);
    }

  else

    { if (s || !norm) 

        { /* Get weights from command line */ 

          do
            { if (winsize==NREC_MAX) 
                exiterror(BADEXPR,argv[0],"too many weights",0);
              w[winsize++]=sgetnum(wspec,&numerr); 
              if (numerr) exiterror(BADNUM,argv[0],wspec,0); 
              wspec=s;
              if (wspec) s=strchr(wspec,',');
              if (s) *(s++)='\0';
            }
          while (wspec); 
        }

      else

        { /* Get window size from command line */

          num=sgetnum(wspec,&numerr);
          if (numerr) exiterror(BADNUM,argv[0],s,0);
          winsize=floor(num);
          if (winsize<1 || winsize!=num) exiterror(BADEXPR,argv[0],wspec,0);
          for (i=0;i<winsize;i++) w[i]=1; 
        }
    }

  num=(float)(winsize+1)/2;
  if (num!=(int)num)
    exiterror(SPECIAL,argv[0],"window size must be odd",0);

  if (norm)
    { sum=0;
      for (i=0;i<winsize;i++) sum+=w[i];
      if (sum==0) exiterror(SPECIAL,argv[0],"sum of weights is zero",0);
      for (i=0;i<winsize;i++) w[i]=w[i]/sum;
    }

  /* END OF WINDOW SPECIFICATION PROCESSING */

  if (strcmp(argv[INFILE],"-")==0) infile=stdin;
  else if ((infile=fopen(argv[INFILE],"r"))==NULL)
    exiterror(OPENINERR,argv[0],argv[INFILE],0);

  status=readvar(infile,varlist,invarname,&ninvar,&ttb,&lineno);
  if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);

  j=VOFFSET;
  found=0;
  while(j<argc && !(found=(strcmp(argv[j],"by")==0))) j++;
  if (found)
    { byoffset=j+1;
      nbyvar=argc-byoffset;
      nnewvar=byoffset-VOFFSET-1;
    }
  else
    { nbyvar=0;
      nnewvar=argc-VOFFSET;
    }
  for (i=0;i<nbyvar;i++) byvarname[i]=argv[byoffset+i];

  if (nbyvar!=0)
    { if (indexlist(nbyvar,byvarname,ninvar,invarname,byvar)<nbyvar)
        { for (i=0;byvar[i]>=0;i++);
          exiterror(VARERR,argv[0],byvarname[i],0);
        }
    }

  nreqvar=0;
  for (i=0;i<ninvar;i++)
    { j=0;
      found=0;
      while (j<nbyvar && !(found=(byvar[j]==i))) j++;
      if (!found)
        { reqvar[nreqvar]=i;
          reqvarname[nreqvar]=invarname[i];
          nreqvar++;
        }
    }
  
  if (ninvar+nnewvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  for (i=0;i<nnewvar;i++)
    { newvarname[i]=argv[i+VOFFSET];
      s=strchr(newvarname[i],'=');
      if (s==NULL || !*(s+1)) exiterror(BADEXPR,argv[0],newvarname[i],0);
      else *(s++)='\0';
      oldvarname[i]=s;
    }
  if (nnewvar>0)
    { if (indexlist(nnewvar,oldvarname,ninvar,invarname,oldvar)<nnewvar)
        { for (i=0;oldvar[i]>=0;i++);
          exiterror(VARERR,argv[0],oldvarname[i],0);
        }
    }

  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else
    { if (ttb) fname=ttbname(argv[OUTFILE],outfname);
      else fname=argv[OUTFILE];
      if ((outfile=fopen(fname,"w"))==NULL)
        exiterror(OPENOUTERR,argv[0],fname,0);
    }

  j=0;
  for (i=0;i<nbyvar;i++) outvarname[j++]=argv[byoffset+i];
  for (i=0;i<nreqvar;i++) outvarname[j++]=reqvarname[i];
  for (i=0;i<nnewvar;i++) outvarname[j++]=newvarname[i];
  status=writevar(outfile,ttb,j,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  outspec.outfile=outfile;
  outspec.ttb=ttb;
  outspec.nbyvar=nbyvar;
  outspec.byvar=byvar;
  outspec.nreqvar=nreqvar;
  outspec.reqvar=reqvar;
  outspec.nnewvar=nnewvar;
  dbuff.fieldlist=NULL;
  dbuff.field=NULL;
  outspec.dbuff=&dbuff;

  /* Find true size of window and offset of leading and trailing ends  
     from mid point: nright,nleft are the number of weights to the right and
     left of centre. If the true window does not span the mid-point then
     one or other will be negative. */ 
 
  for (i=0; i<winsize && w[i]==0; i++) zleft++;
  for (i=winsize-1; i>=0 && w[i]==0; i--) zright++;
  nwin=winsize-zleft-zright;
  if (nwin<0) nwin=0;
  if (nwin==0) 
    { nleft=0;
      nright=0;
    }
  else
    { winmid=floor(winsize/2);
      nleft=winmid-zleft;
      nright=winmid-zright;
    }

  /* Length of holding list is the true window size modified to include 
     the mid point. This ensures access to the current data point at the same 
     time as the data required for calculation of the result. */

  nhold=nwin;
  if (nhold==0) nhold=1;
  else if (nleft<0) nhold=nhold-nleft;
  else if (nright<0) nhold=nhold-nright;

  while (!fin && (status=
         mreaddata(infile,ttb,ninvar,nbyvar,byvar,0,NULL,&dnewrec,&lineno))<=OK)

    { if (status==EOF) fin=1;

      if (!first &&
          (fin || (datacomp(nbyvar,dnewrec,byvar,dhead,byvar,NUMCOMP))!=0))

        { /* Output any leading results for last group not already done */

          if (newgroup && nleft<0 && count>=nwin)
            { dwin=dhead;
              n=count-nwin+1;
              for (i=0;i<-nleft;i++)
                { if (i<n) 
                    { status=
                        writeresult(dwin,oldvar,nwin,w,zleft,NULL,&outspec);
                      dwin=dwin->next;
                    }
                  else 
                    status=writeblank(dhead,NULL,&outspec);
                  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
                }
            }

          /* Output any trailing data records for last group 
             or all records if none already done */

          if (newgroup)
           { dout=dhead;
             n=count;
           }
          else 
           { dout=dout->next;
             n=nright;
           }

          for (i=0;i<n;i++)
            { status=writeblank(dhead,dout,&outspec);
              if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
              dout=dout->next;
            }

          /* Output any trailing results for last group */

          if (nright<0 && count>=nwin)
            { if (newgroup) dwin=dhead;
              else dwin=dwin->next;
              n=count-nwin+1;
              for (i=0;i<-nright;i++)
                { if (i<-nright-n) 
                    status=writeblank(dhead,NULL,&outspec);
                  else
                    { status=
                        writeresult(dwin,oldvar,nwin,w,zleft,NULL,&outspec);
                      dwin=dwin->next;
                    }
                  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
                }
            }

          freelist(&dhead);

          newgroup=1;
          count=0;
        }

      if (!fin)

        { first=0;

          if (!ttb) for (i=0;i<nnewvar;i++)
          dnewrec->x[oldvar[i]]=sgetnum(dnewrec->field[oldvar[i]],&numerr);

          if (dhead==NULL) dhead=dnewrec;
          else d->next=dnewrec;
          d=dnewrec;

          if (count<nhold) count++;
          if (count==nhold)

            { if (newgroup)

                { /* Output any leading data records */

                  dout=dhead;
                  for (i=0;i<nleft;i++)

                    { status=writeblank(dhead,dout,&outspec);
                      if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
                      dout=dout->next;
                    }

                  /* Output any leading results */

                  dwin=dhead;
                  for (i=0;i<-nleft;i++)
                    { status=
                        writeresult(dwin,oldvar,nwin,w,zleft,NULL,&outspec);
                      if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
                      dwin=dwin->next;
                    }

                  newgroup=0;
                }

              else 

                { dout=dout->next;
                  dwin=dwin->next;

                  djunk=dhead;
                  dhead=dhead->next;
                  djunk->next=NULL;
                  freelist(&djunk);
                }

              status=writeresult(dwin,oldvar,nwin,w,zleft,dout,&outspec);
              if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
            }
        }
    }

  if (!fin) exiterror(status,argv[0],argv[INFILE],lineno);

  fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}


int writeresult (struct datarec *dwin, int var[], int nwin, double w[], 
                 int zleft, struct datarec *dout, struct outspecrec *p)

{ struct datarec *d;
  int valid,i,j,status;
  double val,sum;

  for (i=0; i<p->nnewvar; i++)

    { valid=1;
      sum=0;
      d=dwin;

      for (j=0; j<nwin && valid; j++)
        { val=d->x[var[i]];
          if (w[j+zleft]!=0)
            { if (val==NODATA) valid=0;
              else sum+=w[j+zleft]*val;
            }
          d=d->next;
        }
      if (valid) p->dbuff->x[i]=sum;
      else p->dbuff->x[i]=NODATA;
    }

  status=writedata(p->outfile,p->ttb,dwin,p->nbyvar,p->byvar,NULL,NULL,
                   DEFAULT,DEFAULT,0);
  if (status!=OK) return status;
  status=writedata(p->outfile,p->ttb,dout,p->nreqvar,p->reqvar,NULL,NULL,
                   DEFAULT,DEFAULT,0);
  if (status!=OK) return status;
  status=writedata(p->outfile,p->ttb,p->dbuff,p->nnewvar,NULL,NULL,NULL,
                   DEFAULT,DEFAULT,1);
  return status; 
}


int writeblank (struct datarec *dby, struct datarec *d, struct outspecrec *p)

{ int status;

  status=writedata(p->outfile,p->ttb,dby,p->nbyvar,p->byvar,NULL,NULL,
                   DEFAULT,DEFAULT,0);
  if (status!=OK) return status;
  status=writedata(p->outfile,p->ttb,d,p->nreqvar,p->reqvar,NULL,NULL,
                   DEFAULT,DEFAULT,0);
  if (status!=OK) return status;
  status=writedata(p->outfile,p->ttb,NULL,p->nnewvar,NULL,NULL,NULL,
                   DEFAULT,DEFAULT,1);
  return status;
}

