/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* tabjoin.c - John Hemmings (Oct 2000) */

/* v1.01 - added status checks before order checks to avoid UNSORTED errors
           at EOF */  

#include "toolbox.h"

#define FILE1 1
#define FILE2 2
#define OUTFILE 3
#define VOFFSET 4


main (int argc, char *argv[])

{ FILE *infile1,*infile2,*outfile;
  char outfname[FILENAME_MAX+1];
  char *fname1,*fname2,*fname;
  char varlist1[LINE_MAX+1];
  char varlist2[LINE_MAX+1];
  char *invarname1[NFIELD_MAX],*invarname2[NFIELD_MAX],*byvarname[NFIELD_MAX],
       *outvarname[NFIELD_MAX];
  int byvar1[NFIELD_MAX],byvar2[NFIELD_MAX],var1[NFIELD_MAX],var2[NFIELD_MAX],
      commvar[NFIELD_MAX];
  static double x1[NFIELD_MAX];
  static char fieldlist1[LINE_MAX+1];
  static char *field1[NFIELD_MAX];
  struct datarec d1={x1,fieldlist1,field1,NULL};
  static double x2[NFIELD_MAX];
  static char fieldlist2[LINE_MAX+1];
  static char *field2[NFIELD_MAX];
  struct datarec d2={x2,fieldlist2,field2,NULL};
  static double x3[NFIELD_MAX];
  static char fieldlist3[LINE_MAX+1];
  static char *field3[NFIELD_MAX];
  struct datarec dprev={x3,fieldlist3,field3,NULL};
  struct datarec *d,*dnew;
  int byoffset,ninvar1,ninvar2,nbyvar,nvar1,nvar2,noutvar,found,
      status,status1,status2,lineno1,lineno2,ttb1,ttb2,ttb,c,i,j,count,
      nonmatch1,nonmatch2;

  version("1.01",VHDR,argc,argv);

  if (argc<4)
    { fprintf(stderr,
        "Usage: %s [=]infile1 [=]infile2 outfile [keyvar1 ...]\n",argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (argc>VOFFSET)
    { if (strcmp(argv[VOFFSET],"by")==0) byoffset=VOFFSET+1;
      else byoffset=VOFFSET;
      nbyvar=argc-byoffset;
    }
  else nbyvar=0;

  if (nbyvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);
  for (i=0;i<nbyvar;i++) byvarname[i]=argv[byoffset+i];

  /* Get file 1 variables and index key */

  if (*argv[FILE1]=='=')
    { nonmatch1=1;
      fname1=argv[FILE1]+1;
    }
  else 
    { nonmatch1=0;
      fname1=argv[FILE1];
    }

  if (strcmp(fname1,"-")==0) infile1=stdin;
  else if ((infile1=fopen(fname1,"r"))==NULL)
    exiterror(OPENINERR,argv[0],fname1,0);

  status=readvar(infile1,varlist1,invarname1,&ninvar1,&ttb1,&lineno1);
  if (status!=OK) exiterror(status,argv[0],fname1,lineno1);

  if (nbyvar>0)
    { found=indexlist(nbyvar,byvarname,ninvar1,invarname1,byvar1);
      if (found<nbyvar)
        { for (i=0;byvar1[i]>=0;i++);
          exiterror(VARERR,argv[0],byvarname[i],0);
        }
    }

  /* Get file 2 variables and index key */

  if (*argv[FILE2]=='=')
    { nonmatch2=1;
      fname2=argv[FILE2]+1;
    }
  else
    { nonmatch2=0;
      fname2=argv[FILE2];
    }

  if (strcmp(fname2,"-")==0) infile2=stdin;
  else if ((infile2=fopen(fname2,"r"))==NULL)
    exiterror(OPENINERR,argv[0],fname2,0);

  status=readvar(infile2,varlist2,invarname2,&ninvar2,&ttb2,&lineno2);
  if (status!=OK) exiterror(status,argv[0],fname2,lineno2);

  if (nbyvar>0)
    { found=indexlist(nbyvar,byvarname,ninvar2,invarname2,byvar2);
      if (found<nbyvar)
        { for (i=0;byvar2[i]>=0;i++);
          exiterror(VARERR,argv[0],byvarname[i],0);
        }
    }

  /* If no join key specified find common variables */

  if (nbyvar==0)
    { indexlist(ninvar1,invarname1,ninvar2,invarname2,commvar);
      for (i=0;i<ninvar1;i++)
        { if (commvar[i]>=0)
            { byvar1[nbyvar]=i;
              byvar2[nbyvar]=commvar[i];
              byvarname[nbyvar]=invarname1[i];
              nbyvar++;
            }
        }
    }

  /* Index non-key variables */

  nvar1=0;
  for (i=0;i<ninvar1;i++)
    { found=0;
      j=0;
      while (j<nbyvar && !(found=(byvar1[j++]==i)));
      if (!found)
        { var1[nvar1]=i;
          nvar1++;
        }
    }

  nvar2=0;
  for (i=0;i<ninvar2;i++)
    { found=0;
      j=0;
      while (j<nbyvar && !(found=(byvar2[j++]==i)));
      if (!found)
        { var2[nvar2]=i;
          nvar2++;
        }
    }

  /* Set output variables */

  noutvar=nbyvar+nvar1+nvar2;
  if (noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  ttb=(ttb1 && ttb2);
  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else
    { if (ttb) fname=ttbname(argv[OUTFILE],outfname);
      else fname=argv[OUTFILE];
      if ((outfile=fopen(fname,"w"))==NULL)
        exiterror(OPENOUTERR,argv[0],fname,0);
    }

  j=0;
  for (i=0;i<nbyvar;i++) outvarname[j++]=byvarname[i];
  for (i=0;i<nvar1;i++) outvarname[j++]=invarname1[var1[i]];
  for (i=0;i<nvar2;i++) outvarname[j++]=invarname2[var2[i]];
  status=writevar(outfile,ttb,noutvar,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  /* Process data */

  status1=readdata(infile1,ttb1,ninvar1,nbyvar,byvar1,0,NULL,&d1,&lineno1);
  status2=readdata(infile2,ttb2,ninvar2,nbyvar,byvar2,0,NULL,&d2,&lineno2);

  while (status1==OK && status2==OK)

    { c=datacomp(nbyvar,&d1,byvar1,&d2,byvar2,NUMCOMP);
      if (c<0) 

        { if (nonmatch1)   
            { status=writedata(outfile,ttb,&d1,nbyvar,byvar1,NULL,NULL,DEFAULT,
                               DEFAULT,0);
              if (status==OK)
                status=writedata(outfile,ttb,&d1,nvar1,var1,NULL,NULL,
                                 DEFAULT,DEFAULT,0);
              if (status==OK)
                status=writedata(outfile,ttb,NULL,nvar2,NULL,NULL,NULL,DEFAULT,
                                 DEFAULT,1);
              if (status!=OK)
                exiterror(status1,argv[0],argv[OUTFILE],0);
            }

          copydata(&dprev,&d1,nbyvar,byvar1); 
          status1=
            readdata(infile1,ttb1,ninvar1,nbyvar,byvar1,0,NULL,&d1,&lineno1);
          if (status1==OK && datacomp(nbyvar,&d1,byvar1,&dprev,NULL,NUMCOMP)<0)
            status1=UNSORTED;
        }
 
      else if (c>0)

        { if (nonmatch2)

            { status=writedata(outfile,ttb,&d2,nbyvar,byvar2,NULL,NULL,DEFAULT,
                               DEFAULT,0);
              if (status==OK)
                status=writedata(outfile,ttb,NULL,nvar1,NULL,NULL,NULL,DEFAULT,
                                 DEFAULT,0);
              if (status==OK)
                status=writedata(outfile,ttb,&d2,nvar2,var2,NULL,NULL,
                                 DEFAULT,DEFAULT,1);
              if (status!=OK)
                exiterror(status1,argv[0],argv[OUTFILE],0);
            }

          copydata(&dprev,&d2,nbyvar,byvar2); 
          status2=
            readdata(infile2,ttb2,ninvar2,nbyvar,byvar2,0,NULL,&d2,&lineno2);
          if (status2==OK && datacomp(nbyvar,&d2,byvar2,&dprev,NULL,NUMCOMP)<0)
            status2=UNSORTED;
        }

      else

        { /* Get and store by group from file 2 */ 

          count=1;
          d=&d2;
          do 
            { status2=mreaddata(infile2,ttb2,ninvar2,nbyvar,byvar2,0,NULL,&dnew,
                                &lineno2); 
              if (status2==OK)
                { c=datacomp(nbyvar,dnew,byvar2,&d1,byvar1,NUMCOMP);
                  if (c==0) 
                    { if (++count>NREC_MAX) 
                        exiterror(RECLIMIT,argv[0],fname2,lineno2); 
                      d->next=dnew;
                      d=dnew;
                    } 
                  else if (c<0) status2=UNSORTED;
                }
            }
          while (status2==OK && c==0);

          /* Join all matching file 1 records with stored group */ 

          do

            { d=&d2;
              do
                { status=writedata(outfile,ttb,&d1,nbyvar,byvar1,NULL,NULL,
                                   DEFAULT,DEFAULT,0);
                  if (status==OK)
                    status=writedata(outfile,ttb,&d1,nvar1,var1,NULL,NULL,
                                     DEFAULT,DEFAULT,0);
                  if (status==OK)
                    status=writedata(outfile,ttb,d,nvar2,var2,NULL,NULL,DEFAULT,
                                     DEFAULT,1);
                  if (status!=OK) 
                    exiterror(status1,argv[0],argv[OUTFILE],0); 
                  d=d->next;
                }
              while (d!=NULL);

              status1=readdata(infile1,ttb1,ninvar1,nbyvar,byvar1,0,NULL,&d1,
                               &lineno1);
              if (status1==OK) 
                { c=datacomp(nbyvar,&d1,byvar1,&d2,byvar2,NUMCOMP);
                  if (c<0) status1=UNSORTED;
                }
            }

          while (status1==OK && c==0);

          /* Copy next file2 record to buffer */

          if (status2==OK) 
            { freelist(&d2.next); 
              copydata(&d2,dnew,ninvar2,NULL);
              freelist(&dnew); 
            }
        }
    }

  if (nonmatch1)
    { while (status1==OK)

        { status=writedata(outfile,ttb,&d1,nbyvar,byvar1,NULL,NULL,DEFAULT,
                           DEFAULT,0);
          if (status==OK)
            status=writedata(outfile,ttb,&d1,nvar1,var1,NULL,NULL,
                             DEFAULT,DEFAULT,0);
          if (status==OK)
            status=writedata(outfile,ttb,NULL,nvar2,NULL,NULL,NULL,DEFAULT,
                             DEFAULT,1);
          if (status!=OK)
            exiterror(status1,argv[0],argv[OUTFILE],0);

          copydata(&dprev,&d1,nbyvar,byvar1);
          status1=
            readdata(infile1,ttb1,ninvar1,nbyvar,byvar1,0,NULL,&d1,&lineno1);
          if (status1==OK && datacomp(nbyvar,&d1,byvar1,&dprev,NULL,NUMCOMP)<0)
            status1=UNSORTED;
        }
    }
  
  if (nonmatch2)
    { while (status2==OK)

        { status=writedata(outfile,ttb,&d2,nbyvar,byvar2,NULL,NULL,DEFAULT,
                           DEFAULT,0);
          if (status==OK)
            status=writedata(outfile,ttb,NULL,nvar1,NULL,NULL,NULL,DEFAULT,
                             DEFAULT,0);
          if (status==OK)
            status=writedata(outfile,ttb,&d2,nvar2,var2,NULL,NULL,
                             DEFAULT,DEFAULT,1);
          if (status!=OK)
            exiterror(status1,argv[0],argv[OUTFILE],0);
          
          copydata(&dprev,&d2,nbyvar,byvar2);
          status2=
            readdata(infile2,ttb2,ninvar2,nbyvar,byvar2,0,NULL,&d2,&lineno2);
          if (status2==OK && datacomp(nbyvar,&d2,byvar2,&dprev,NULL,NUMCOMP)<0)
            status2=UNSORTED;
        }

    }
  
  if (status1!=OK && status1!=EOF) 
    exiterror(status1,argv[0],fname1,lineno1);
  if (status2!=OK && status2!=EOF) 
    exiterror(status2,argv[0],fname2,lineno2);

  fclose(infile1);
  fclose(infile2);
  fclose(outfile);
  return EXIT_SUCCESS;
}

