/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabmerge.c - John Hemmings (Oct 2000) */

/* v1.01 - Dec 2000 - fixed bug causing wrong variable to be use in sortcomp
           when output variables (reqvar) are ordered differently in different
           input files

   v1.02 - Jul 2006 - fixed bug affecting count of number of open files +
           removed redundant 'static' qualifier from declarations in 'main' 
*/


#include "toolbox.h"

#define MAXFILES 200 

struct filerec {
  char *name;
  FILE *ptr;
  long pos;
  int ttb;
  int nvar;
  int nstrvar;
  int reqvar[NFIELD_MAX];
  struct datarec d;
  int lineno;
  int status;
};


int main (int argc, char *argv[])

{ struct filerec infile[MAXFILES],*f;
  FILE *outfile;
  char outfname[FILENAME_MAX+1];
  char *fname;
  char invarlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX];
  char *reqvarname[NFIELD_MAX];
  char *sortvarname[NFIELD_MAX];
  char *auxvarname[NFIELD_MAX];
  char *outvarname[NFIELD_MAX];
  char reqvarlist[LINE_MAX+1];
  struct sortrec sortspec;
  int auxvar[NFIELD_MAX];
  double x[MAXFILES][NFIELD_MAX];
  char fieldlist[MAXFILES][LINE_MAX+1];
  char *field[MAXFILES][NFIELD_MAX];
  double x1[NFIELD_MAX];
  char fieldlist1[LINE_MAX+1];
  char *field1[NFIELD_MAX];
  struct datarec dbuff={x1,fieldlist1,field1,NULL};
  double x2[NFIELD_MAX];
  char *field2[NFIELD_MAX];
  struct datarec dnew={x2,NULL,field2,NULL};
  int voffset,ioutfname,nreqvar,nauxvar=0,found,nfiles,nopen,nfin,status,i,j,
      imin=0,just,ttbout,dummy;

  version("1.02",VHDR,argc,argv);

  if (argc>=5)

    { for (i=1;i<argc;i++)
        if (strpbrk(argv[i],WHITESPACE)!=NULL) 
          exiterror(BADARG,argv[0],argv[i],0);

      j=0;
      found=0;
      while(j<argc && !(found=(strcmp(argv[j],"on")==0))) j++;
      if (found)
        { voffset=j+1;
          sortspec.nvar=argc-voffset;
        }
      else sortspec.nvar=0;

      nfiles=voffset-3;
    }

  if (argc<5 || sortspec.nvar<1 || nfiles<1)
    { fprintf(stderr,
"Usage: %s infile1 [infile2 ...] outfile on [-][@]var1 [[-][@]var2 ...]\n",
              argv[0]);
      exit(EXIT_FAILURE);
    }

  if (nfiles>MAXFILES) exiterror(FILELIMIT,argv[0],argv[MAXFILES+1],0);
  found=0;
  for (i=0;i<nfiles;i++)
    { if (strcmp(argv[i+1],"-")==0)
        { if (found) 
            exiterror(SPECIAL,argv[0],"multiple standard input streams",0);
          else found=1;
        }
      infile[i].name=argv[i+1];
      infile[i].d.x=x[i];
      infile[i].d.fieldlist=fieldlist[i];
      infile[i].d.field=field[i];
    }

  /* Determine required variables and format from first file */

  f=&infile[0];

  if (strcmp(f->name,"-")==0) f->ptr=stdin;
  else if ((f->ptr=fopen(f->name,"r"))==NULL)
    exiterror(OPENINERR,argv[0],f->name,0);
  nopen=1;

  status=readvar(f->ptr,reqvarlist,reqvarname,&f->nvar,&f->ttb,&f->lineno);
  if (status!=OK) exiterror(status,argv[0],f->name,f->lineno);

  nreqvar=f->nvar;
  indexlist(nreqvar,reqvarname,nreqvar,reqvarname,f->reqvar);

  ttbout=f->ttb;
  if (ttbout) f->nstrvar=0;
  else f->nstrvar=nreqvar;

  /* Set up comparison specification for 'on' variables */

  for (i=0;i<sortspec.nvar;i++)
    { sortvarname[i]=argv[voffset+i];
      if (*sortvarname[i]=='-')
        { sortspec.sense[i]=-1;
          sortvarname[i]++;
        }
      else sortspec.sense[i]=1;
      if (*sortvarname[i]=='@')
        { if (ttbout) exiterror(BADEXPR,argv[0],sortvarname[i],0);
          sortspec.ctype[i]=STRCOMP;
          sortvarname[i]++;
        }
      else sortspec.ctype[i]=NUMCOMP;
    }

  if (indexlist(sortspec.nvar,sortvarname,nreqvar,reqvarname,sortspec.var)<
      sortspec.nvar)
    { for (i=0;sortspec.var[i]>=0;i++);
      exiterror(VARERR,argv[0],sortvarname[i],0);
    }

  for (i=0;i<nreqvar;i++)
    { found=0;
      j=0;
      while (j<sortspec.nvar && !(found=(sortspec.var[j]==i))) j++;
      if (!found)
        { auxvar[nauxvar]=i;
          auxvarname[nauxvar]=reqvarname[i];
          nauxvar++;
        }
    }

  nfin=nfiles;
  for (i=0;i<nfiles;i++)

    { f=&infile[i];
      if (i>0)

        { if (nopen<FOPEN_MAX) nopen++;
          else
            { j=i-1; 
              while (infile[j].ptr==NULL || infile[j].ptr==stdin) j--;
              fclose(infile[j].ptr);
              infile[j].ptr=NULL;
            }
          if (strcmp(f->name,"-")==0) f->ptr=stdin;
          else if ((f->ptr=fopen(f->name,"r"))==NULL)
            exiterror(OPENINERR,argv[0],f->name,0);

          status=readvar(f->ptr,invarlist,invarname,&f->nvar,&f->ttb,
                         &f->lineno);
          if (status!=OK)
            exiterror(status,argv[0],f->name,f->lineno);

          if (indexlist(nreqvar,reqvarname,f->nvar,invarname,f->reqvar)<
              nreqvar)
            { for (j=0;f->reqvar[j]>=0;j++);
              exiterror(VARERR,argv[0],reqvarname[j],0);
            }
        }

      /* Get first record (convert to required output format if necessary) */

      if (ttbout) f->nstrvar=0;
      else f->nstrvar=nreqvar;

      f->pos=ftell(f->ptr);
      f->status=readdata(f->ptr,f->ttb,f->nvar,0,NULL,f->nstrvar,f->reqvar,
                         &dbuff,&f->lineno);

      if (f->status==OK)
        { if (ttbout && !f->ttb)
            { for (j=0;j<nreqvar;j++)
                dbuff.x[f->reqvar[j]]=sgetnum(dbuff.field[f->reqvar[j]],&dummy);
            }
          copydata(&f->d,&dbuff,nreqvar,f->reqvar);
        }
      else if (f->status==EOF)
        { fclose(f->ptr);
          f->ptr=NULL;
          nopen--;
          nfin--;
        }
      else exiterror(f->status,argv[0],f->name,f->lineno);
    }

  if (sortspec.nvar+nauxvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  ioutfname=i+1;
  if (strcmp(argv[ioutfname],"-")==0) outfile=stdout;
  else
    { if (ttbout) fname=ttbname(argv[ioutfname],outfname);
      else fname=argv[ioutfname];
      if ((outfile=fopen(fname,"w"))==NULL)
        exiterror(OPENOUTERR,argv[0],fname,0);
    }

  j=0;
  for (i=0;i<sortspec.nvar;i++) outvarname[j++]=sortvarname[i];
  for (i=0;i<nauxvar;i++) outvarname[j++]=auxvarname[i];
  status=writevar(outfile,ttbout,j,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[ioutfname],0);

  while (nfin>0)

    { /* Select input file for next output record */

      if (infile[imin].status!=OK) 
        { imin=0;
          while (infile[imin].status!=OK) imin++; 
        }

      for (i=0;i<nfiles;i++)
        if (i!=imin && infile[i].status==OK &&
            sortcomp(&infile[i].d,&infile[imin].d,&sortspec)<0) imin=i;
      f=&infile[imin];

      for (j=0; status==OK && j<sortspec.nvar; j++)
        { if (sortspec.ctype[j]==STRCOMP) just=LEFT;
          else just=DEFAULT;
          status=writedata(outfile,ttbout,&f->d,1,&sortspec.var[j],NULL,
                           NULL,just,DEFAULT,0);
        }
      if (status==OK)
        status=writedata(outfile,ttbout,&f->d,nauxvar,auxvar,NULL,NULL,
                         DEFAULT,DEFAULT,1);
      if (status!=OK) exiterror(status,argv[0],argv[ioutfname],0);

      if (f->ptr==NULL)

        { /* Reopen current input file */
 
          if (nopen==FOPEN_MAX-1) 
            { j=0;
              while (infile[j].ptr==NULL || infile[j].ptr==stdin) j++;
              fclose(infile[j].ptr);
              infile[j].ptr=NULL;
            }
          else nopen++;
          if (strcmp(f->name,"-")==0) f->ptr=stdin;
          else if ((f->ptr=fopen(f->name,"r"))==NULL)
            exiterror(OPENINERR,argv[0],f->name,0);

          if (fseek(f->ptr,f->pos,SEEK_SET)!=0)
            exiterror(SPECIAL,argv[0],"input file has changed",0);
      
          /* Re-read last record to check for changes */
 
          status=readdata(f->ptr,f->ttb,f->nvar,0,NULL,f->nstrvar,f->reqvar,
                          &dbuff,&dummy);
          if (status==OK)
            { if (ttbout && !f->ttb)
                { for (i=0;i<nreqvar;i++)
                    dbuff.x[f->reqvar[i]]=
                      sgetnum(dbuff.field[f->reqvar[i]],&dummy);
                }
            }
          if (status!=OK || 
              datacomp(nreqvar,&dbuff,f->reqvar,&f->d,NULL,
                       (ttbout)?NUMCOMP:STRCOMP)!=0) 
            exiterror(SPECIAL,argv[0],"input file has changed",0);
        }

      /* Get next record from current input file (convert if necessary) */

      f->pos=ftell(f->ptr);
      f->status=readdata(f->ptr,f->ttb,f->nvar,0,NULL,f->nstrvar,f->reqvar,
                         &dbuff,&f->lineno);

      if (f->status==OK)
        { if (ttbout && !f->ttb)
            { for (i=0;i<nreqvar;i++)
                dbuff.x[f->reqvar[i]]=sgetnum(dbuff.field[f->reqvar[i]],&dummy);
            }
          for (i=0;i<nreqvar;i++)
            { dnew.x[i]=dbuff.x[f->reqvar[i]];
              dnew.field[i]=dbuff.field[f->reqvar[i]];
            }
          if (sortcomp(&dnew,&f->d,&sortspec)<0) f->status=UNSORTED;
          copydata(&f->d,&dbuff,nreqvar,f->reqvar);
        }

      if (f->status==EOF) 
        { fclose(f->ptr);
          f->ptr=NULL;
          nopen--;
          nfin--;
        }
      else if (f->status!=OK) 
        exiterror(f->status,argv[0],f->name,f->lineno); 
    }

  fclose(outfile);
  return EXIT_SUCCESS;
}

