/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabsort.c - John Hemmings (Oct 2000) */

/* v1.01 - Aug 2011 - Replaced 'quicksort' with library function 'qsort' 
   v1.02 - Feb 2013 - replaced 'readdata with 'getdata' to handle column data
           record format; added '.ttb' output file name recognition and option
           to suppress header for output to stdout; made 'var1' argument 
           optional; removed special treatment of argument "on"; reduced 
           automatic storage requirements to avoid stack overflow */

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2
#define VOFFSET 3 

int compare (const void *, const void *);
static struct sortrec sortspec;

main (int argc, char *argv[])

{ FILE *infile,*outfile;
  char outfbuff[FILENAME_MAX+1];
  char *outfname;
  char varlist[LINE_MAX+1];
  char **invarname,**sortvarname,**auxvarname,**outvarname;
  int *byvar,*auxvar;
  struct datarec *d[NREC_MAX],*dnew;
  double *x;
  char fieldlist[LINE_MAX+1];
  char **field;
  struct datarec dalt;
  struct readctrlrec readctrl={0,NULL,0,NULL};
  int byoffset,ninvar,nbyvar,nauxvar=0,found,i,j,c,status,first=1,
      fin=0,count=0,just,lineno,ttbin,ttbout;

  version("1.02",VHDR,argc,argv);

  if (argc<3)
    { fprintf(stderr,
              "Usage: %s infile outfile [-][@]var1 ...] [by keyvar1 ...]\n",
	      argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (strcmp(argv[INFILE],"-")==0) infile=stdin;
  else if ((infile=fopen(argv[INFILE],"r"))==NULL)
    exiterror(OPENINERR,argv[0],argv[INFILE],0);

  invarname=(char **)malloc(NFIELD_MAX*sizeof(char *));
  if (!invarname) exiterror(MEMERR,argv[0],NULL,0);
  auxvarname=(char **)malloc(NFIELD_MAX*sizeof(char *));
  if (!auxvarname) exiterror(MEMERR,argv[0],NULL,0);
  outvarname=(char **)malloc(NFIELD_MAX*sizeof(char *));
  if (!outvarname) exiterror(MEMERR,argv[0],NULL,0);
  byvar=(int *)malloc(NFIELD_MAX*sizeof(int));
  if (!byvar) exiterror(MEMERR,argv[0],NULL,0);
  auxvar=(int *)malloc(NFIELD_MAX*sizeof(int));
  if (!auxvar) exiterror(MEMERR,argv[0],NULL,0);

  status=readvar(infile,varlist,invarname,&ninvar,&ttbin,&lineno);
  if (status==OK_INLINE)
    { fclose(infile);
      infile=NULL;
    } 
  else if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);

  j=VOFFSET;
  found=0;
  while(j<argc && !(found=(strcmp(argv[j],"by")==0))) j++;
  if (found)
    { byoffset=j+1;
      nbyvar=argc-byoffset;
      sortspec.nvar=byoffset-VOFFSET-1;
    }
  else
    { nbyvar=0;
      sortspec.nvar=argc-VOFFSET;
    }

  if (sortspec.nvar==0) 

    { x=(double *)malloc(NFIELD_MAX*sizeof(double));
      if (!x) exiterror(MEMERR,argv[0],NULL,0);
      field=(char **)malloc(NFIELD_MAX*sizeof(char *));
      if (!field) exiterror(MEMERR,argv[0],NULL,0);
      dalt.x=x;
      dalt.fieldlist=fieldlist;
      dalt.field=field;
      dalt.next=NULL;
    }

  else

    { sortvarname=(char **)malloc(NFIELD_MAX*sizeof(char *));
      if (!sortvarname) exiterror(MEMERR,argv[0],NULL,0);

      for (i=0;i<sortspec.nvar;i++)
        { sortvarname[i]=argv[VOFFSET+i];
          if (*sortvarname[i]=='-')
	    { sortspec.sense[i]=-1;
	      sortvarname[i]++;
	    }
          else sortspec.sense[i]=1;
          if (*sortvarname[i]=='@')
	    { if (ttbin) exiterror(BADEXPR,argv[0],sortvarname[i],0); 
              sortspec.ctype[i]=STRCOMP;
	      sortvarname[i]++;
            }
          else sortspec.ctype[i]=NUMCOMP;
        }

      if (indexlist(sortspec.nvar,sortvarname,ninvar,invarname,sortspec.var)<
           sortspec.nvar)
        { for (i=0;sortspec.var[i]>=0;i++);
          exiterror(VARERR,argv[0],sortvarname[i],0);
        }
    }

  if (indexlist(nbyvar,&argv[byoffset],ninvar,invarname,byvar)<nbyvar)
    { for (i=0;byvar[i]>=0;i++);
      exiterror(VARERR,argv[0],argv[byoffset+i],0);
    }

  for (i=0;i<ninvar;i++)
    { found=0;
      j=0;
      while (j<nbyvar && !(found=(byvar[j]==i))) j++;
      if (!found)
        { j=0;
          while (j<sortspec.nvar && !(found=(sortspec.var[j]==i))) j++;
          if (!found)
            { auxvar[nauxvar]=i;
              auxvarname[nauxvar]=invarname[i];
              nauxvar++;
            }
	}
    }
 
  if (nbyvar+sortspec.nvar+nauxvar>NFIELD_MAX)
    exiterror(OUTVARERR,argv[0],NULL,0);

  ttbout=ttbin;
  if ((outfile=outfopen(argv[OUTFILE],outfbuff,&outfname,&ttbout))==NULL)
    exiterror(OPENOUTERR,argv[0],outfname,0);

  if (strcmp(argv[OUTFILE],".")!=0) 
    { j=0;
      for (i=0;i<nbyvar;i++) outvarname[j++]=argv[byoffset+i];
      for (i=0;i<sortspec.nvar;i++) outvarname[j++]=sortvarname[i];
      for (i=0;i<nauxvar;i++) outvarname[j++]=auxvarname[i];
      status=writevar(outfile,ttbout,j,outvarname,NULL,DEFAULT);
      if (status!=OK) exiterror(status,argv[0],outfname,0);
    }

  readctrl.nkeyvar=nbyvar;
  readctrl.keyvar=byvar;

  if (sortspec.nvar==0) 

    { while ((status=getdata(ninvar,NULL,infile,ttbin,&readctrl,invarname,
                             &dalt,NULL,&lineno))==OK)
        { status=writedata(outfile,ttbout,&dalt,nbyvar,byvar,NULL,NULL,
                           DEFAULT,DEFAULT,0);
          if (status==OK)
            status=writedata(outfile,ttbout,&dalt,nauxvar,auxvar,NULL,NULL,
                             DEFAULT,DEFAULT,1);
          if (status!=OK) exiterror(status,argv[0],outfname,0);
        }
      
      if (status==EOF) fin=1;
    }

  else

    { while (!fin && (status=getdata(ninvar,NULL,infile,ttbin,&readctrl,
                                     invarname,NULL,&dnew,&lineno))<=OK)

        { if (status==EOF) fin=1;
          else if (count==NREC_MAX)
	    exiterror(RECLIMIT,argv[0],argv[INFILE],lineno);

          if (!first &&
              (fin || datacomp(nbyvar,dnew,byvar,d[0],byvar,NUMCOMP)!=0))

	    { qsort(d,count,sizeof(d[0]),compare);

	      for (i=0;i<count;i++)

                { status=writedata(outfile,ttbout,d[i],nbyvar,byvar,NULL,NULL,
                                   DEFAULT,DEFAULT,0);
                  for (j=0; status==OK && j<sortspec.nvar; j++)
                    { if (sortspec.ctype[j]==STRCOMP) just=LEFT;
                      else just=DEFAULT;
	              status=writedata(outfile,ttbout,d[i],1,&sortspec.var[j],
                                       NULL,NULL,just,DEFAULT,0);
                    }
	          if (status==OK)
	            status=writedata(outfile,ttbout,d[i],nauxvar,auxvar,NULL,
                                     NULL,DEFAULT,DEFAULT,1);
	          if (status!=OK) exiterror(status,argv[0],outfname,0);

                  freelist(&d[i]);
	        }

              count=0;
            }

          if (!fin)
            { d[count++]=dnew;
	      first=0;
            }
        }
    }

  if (!fin) exiterror(status,argv[0],argv[INFILE],lineno);

  if (infile) fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}


int compare (const void *dtest, const void *dref) 

{ return sortcomp(*(struct datarec **)dtest,*(struct datarec **)dref,&sortspec);
}



