/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabstr.c - John Hemmings (Dec 2001) */

/* v1.01 - Aug 2002 - Use "" instead of NODATA_STR in 'getcond' +
           Protect against output of blank strings + 
           Fixed bug in 'IF' processing +
           Allow quotes in strings using '\' +
           New functions 'mirror' 'lenmatch' 'upper' 'lower'
   v1.02 - Nov 2004 - New function 'printf' 
   v1.03 - Aug 2005 - Fixed to avoid passing NULL to fclose 
   v1.04 - May 2006 - Added '.ttb' output file name recognition and
           option to suppress header for output to stdout ('outfile' = '.'
           or can be omitted; infile can also be omitted if no input data) +
           Fixed 'readdata' call for ttb input to force conversion
   v1.05 - Sep 2006 - New functions 'ltrim','rtrim','trim'
   v1.06 - Nov 2013 - 'sprintf' argument compatibility warnings addressed for 
           'len' and 'posof' functions
*/

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2

enum {BADEXPR_COND=-99,BADEXPR_STR1,BADEXPR_STR2,BADEXPR_STR3,BADEXPR_STR4,
      BADEXPR_STR5};

enum {CONST,ID,CAT,AND,OR,SUBSTR,MIRROR,LEN,LENMATCH,POSOF,TR,UPPER,LOWER,
      LTRIM,RTRIM,TRIM,PRINTF,MIN,MAX,IF,EQ,NE,LT,GT,LE,GE,NOT};

struct node {
  int op;
  int index;
  char *svalue;
  struct node *arg;
  struct node *arg2;
  struct node *arg3;
  struct node *next;
};

int parse (char **, char *, int, char * [], struct node **);
int getandexpr (char **, char *, int, char * [], struct node **);
int getcond (char **, char *, int, char * [], struct node **);
int getcatlist (char **, char *, int, char * [], struct node **);
int getstr (char **, char *, int, char * [], struct node **);
char *endquote (char *, char);
char *evalnode (struct node *, char * [], char *);
char *evalcat (struct node *, char * [], char *);
char *evaland (struct node *, char * [], char *);
char *evalor (struct node *, char * [], char *);


int main (int argc, char *argv[])

{ FILE *infile=NULL, *outfile=NULL;
  char outfbuff[FILENAME_MAX+1];
  char *outfnamereq; 
  char *outfname; 
  char varlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX],*newvarname[NFIELD_MAX],*outvarname[NFIELD_MAX];
  static double x[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec d={x,fieldlist,field,NULL};
  struct node *expr[NFIELD_MAX];
  char currexpr[LINE_MAX+1];
  char *s,dlim;
  static char fieldlist1[LINE_MAX+1];
  static char *field1[1];
  struct datarec dnew={NULL,fieldlist1,field1,NULL};
  int ninvar,nnewvar,noutvar,status,i,ttb,ttbout,lineno,
      voffset=3,noinfile=0,nooutfilehead=0;

  version("1.06",VHDR,argc,argv);

  if (argc<2)
    { fprintf (stderr,"Usage: %s infile outfile [var1=expr1 ...]\n",argv[0]);
      fprintf (stderr,"   or: %s [infile] var1=expr1 [...]\n",argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (*argv[INFILE]=='=') 
    { noinfile=1;
      nooutfilehead=1;
      voffset=1;
    }
  else if (argc<3 || *argv[OUTFILE]=='=')
    { nooutfilehead=1;
      voffset=2;
    }

  if (!noinfile && strcmp(argv[INFILE],".")==0) noinfile=1;
  if (!nooutfilehead && strcmp(argv[OUTFILE],".")==0) nooutfilehead=1;

  if (noinfile) ninvar=0;
  else

    { if (strcmp(argv[INFILE],"-")==0) infile=stdin;
      else if ((infile=fopen(argv[INFILE],"r"))==NULL)
        exiterror(OPENINERR,argv[0],argv[INFILE],0);

      status=readvar(infile,varlist,invarname,&ninvar,&ttb,&lineno);
      if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);
    }
  
  nnewvar=argc-voffset;
  noutvar=ninvar+nnewvar;
  if (noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  currexpr[LINE_MAX]='\0';
  for (i=0;i<nnewvar;i++)
    { strncpy(currexpr,argv[i+voffset],LINE_MAX);
      newvarname[i]=argv[i+voffset];
      s=strchr(newvarname[i],'=');
      if (s==NULL) s="";
      else *(s++)='\0';
      if (!nooutfilehead && !validvname(newvarname[i]))
        exiterror(BADEXPR,argv[0],newvarname[i],0);
      status=parse(&s,&dlim,ninvar,invarname,&expr[i]);
      if (status==VARERR) exiterror(VARERR,argv[0],s,0);
      if (status==BADFUNC) exiterror(BADFUNC,argv[0],s,0);
      if (status<0) exiterror(BADEXPR,argv[0],currexpr,0);
      if (status!=OK) exiterror(status,argv[0],currexpr,0);
      if (dlim!='\0') exiterror(BADEXPR,argv[0],currexpr,0);
    }

  if (nooutfilehead) outfnamereq=".";
  else outfnamereq=argv[OUTFILE];
  ttbout=0;
  if ((outfile=outfopen(outfnamereq,outfbuff,&outfname,&ttbout))==NULL)
    exiterror(OPENOUTERR,argv[0],outfname,0);

  if (!nooutfilehead)
    { for (i=0;i<ninvar;i++) outvarname[i]=invarname[i];
      for (i=0;i<nnewvar;i++) outvarname[ninvar+i]=newvarname[i];
      status=writevar(outfile,ttbout,noutvar,outvarname,NULL,DEFAULT);
      if (status!=OK) exiterror(status,argv[0],outfname,0);
    }

  while ((!noinfile && 
          (status=readdata(infile,ttb,ninvar,0,NULL,ninvar,NULL,&d,&lineno))
           ==OK) || (noinfile && status==OK))

    { if (!nooutfilehead) 
        { status=writedata(outfile,ttbout,&d,ninvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,nnewvar==0?1:0);
          if (status!=OK) exiterror(status,argv[0],outfname,0);
        }

      for (i=0;i<nnewvar;i++)
        { *dnew.field=evalnode(expr[i],d.field,dnew.fieldlist);
          if (*dnew.field && ! **dnew.field) *dnew.field=NULL; 
          status=writedata(outfile,ttbout,&dnew,1,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,i==nnewvar-1?1:0);
          if (status!=OK) exiterror(status,argv[0],outfname,0);
        }

      if (noinfile) status=EOF;
    }

  if (status!=EOF) exiterror(status,argv[0],argv[INFILE],lineno);

  if (infile) fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}


int parse (char **s, char *dlim, int nvar, char *varname[], struct node
**head)

{ struct node *newnode,*curr;
  int status;

  *dlim='\0';

  if (**s=='\0')
    { *head=NULL;
      return OK;
    }

  status=getandexpr(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='|')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=OR;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getandexpr(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='|');

  return OK;
}


int getandexpr (char **s, char *dlim, int nvar, char *varname[],
                struct node **head)

{ struct node *curr,*newnode;
  int status;

  *dlim='\0';

  status=getcond(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='&')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=AND;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getcond(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='&');

  return OK;
}


int getcond (char **s, char *dlim, int nvar, char *varname[],
             struct node **head)

{ struct node *newnode;
  int status;

  *dlim='\0';

  status=getcatlist(s,dlim,nvar,varname,&newnode); 
  if (status!=OK) return status;
  if (*dlim!='=' && *dlim!='!' && *dlim!='<' && *dlim!='>')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  if (*dlim=='=') (*head)->op=EQ;
  else if (*dlim=='!')
    { if (**s=='=')
        { (*head)->op=NE;
          (*s)++;
        }
      else return BADEXPR_COND;
    }
  else if (*dlim=='<')
    { if (**s=='=')
        { (*head)->op=LE;
          (*s)++;
        }
      else (*head)->op=LT;
    }
  else if (*dlim=='>')
    { if (**s=='=')
        { (*head)->op=GE;
          (*s)++;
        }
      else (*head)->op=GT;
    }
  (*head)->arg=newnode;

  if (((*head)->op==EQ || (*head)->op==NE) &&
      (**s==')' || **s==',' || **s=='\0'))
    { newnode=malloc(sizeof(struct node));
      if (newnode==NULL) return MEMERR;
      newnode->op=CONST;
      newnode->svalue="";
      *dlim=**s;
      (*s)++;
    }
  else
    { status=getcatlist(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
    }
  (*head)->arg2=newnode;

  return OK;
}


int getcatlist (char **s, char *dlim, int nvar, char *varname[],
                struct node **head)

{ struct node *curr,*newnode;
  int status;

  *dlim='\0';

  status=getstr(s,dlim,nvar,varname,&newnode); 
  if (status!=OK) return status;
  if (*dlim!='+' && *dlim!='\'' && *dlim!='\"')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=CAT;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getstr(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='+' || *dlim=='\'' || *dlim=='\"');

  return OK;
}


int getstr (char **s, char *dlim, int nvar, char *varname[],
            struct node **f)

{ struct node *newnode;
  char *p,quote='\0';
  int status,i,found;

  if (*dlim=='\0' && **s=='+') (*s)++;

  if (*dlim=='\'' || *dlim=='\"') quote=*dlim;
  else if (**s=='\'' || **s=='\"') 
    { quote=**s; 
      (*s)++; 
    }

  if (quote)

    { p=endquote(*s,quote);
      if (p)
        { *p='\0';
          *f=malloc(sizeof(struct node));
          if (*f==NULL) return MEMERR;
          (*f)->op=CONST;
          (*f)->svalue=*s;
          p++;
          if (strchr("+|&=!<>),",*p))
            { *dlim=*p;
              *s=p+1;
            }
          else
            { *dlim='+';
              *s=p;
            }
        }
      else return BADEXPR_STR1;
    }

  else

    { p=strpbrk(*s,"+()|&=!<>,\'\"");

      if (p && *p=='(')

        { if (p==*s)
            { *s=p+1;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              *f=newnode;
            }

          else if (isalpha(**s))

            { *p='\0';
              *f=malloc(sizeof(struct node));
              if (*f==NULL) return MEMERR;
              if (strcmp(*s,"substr")==0) (*f)->op=SUBSTR;
              else if (strcmp(*s,"mirror")==0) (*f)->op=MIRROR;
              else if (strcmp(*s,"len")==0) (*f)->op=LEN;
              else if (strcmp(*s,"lenmatch")==0) (*f)->op=LENMATCH;
              else if (strcmp(*s,"posof")==0) (*f)->op=POSOF;
              else if (strcmp(*s,"tr")==0) (*f)->op=TR;
              else if (strcmp(*s,"upper")==0) (*f)->op=UPPER;
              else if (strcmp(*s,"lower")==0) (*f)->op=LOWER;
              else if (strcmp(*s,"ltrim")==0) (*f)->op=LTRIM;
              else if (strcmp(*s,"rtrim")==0) (*f)->op=RTRIM;
              else if (strcmp(*s,"trim")==0) (*f)->op=TRIM;
              else if (strcmp(*s,"printf")==0) (*f)->op=PRINTF;
              else if (strcmp(*s,"min")==0) (*f)->op=MIN;
              else if (strcmp(*s,"max")==0) (*f)->op=MAX;
              else if (strcmp(*s,"if")==0) (*f)->op=IF;
              else if (strcmp(*s,"not")==0) (*f)->op=NOT;
              else return BADFUNC;

              *s=p+1;
              if ((*f)->op==LENMATCH || (*f)->op==POSOF || 
                  (*f)->op==RTRIM || (*f)->op==LTRIM || (*f)->op==TRIM ||
                  (*f)->op==PRINTF ||
                  (*f)->op==MIN || (*f)->op==MAX)
                { status=parse(s,dlim,nvar,varname,&newnode);
                  if (status!=OK) return status;
                  if (*dlim!=',') return BADEXPR_STR2;
                  (*f)->arg=newnode;
                  status=parse(s,dlim,nvar,varname,&newnode);
                  if (status!=OK) return status;
                  (*f)->arg2=newnode;
                }
              else if ((*f)->op==IF || (*f)->op==SUBSTR || (*f)->op==TR)
                { status=parse(s,dlim,nvar,varname,&newnode);
                  if (status!=OK) return status;
                  if (*dlim!=',') return BADEXPR_STR3;
                  (*f)->arg=newnode;
                  status=parse(s,dlim,nvar,varname,&newnode);
                  if (status!=OK) return status;
                  (*f)->arg2=newnode;
                  if (*dlim!=',') (*f)->arg3=NULL;
                  else
                    { status=parse(s,dlim,nvar,varname,&newnode);
                      if (status!=OK) return status;
                      (*f)->arg3=newnode;
                    }
                }
              else
                { status=parse(s,dlim,nvar,varname,&newnode);
                  if (status!=OK) return status;
                  (*f)->arg=newnode;
                }
            }

          else return BADEXPR_STR4;

          if (*dlim!=')') return BADEXPR_STR5;

          if (strchr("+|&=!<>)",**s)) 
            { *dlim=**s;
              (*s)++;
            }
          else *dlim='+';
        }

      else

        { *f=malloc(sizeof(struct node));
          if (*f==NULL) return MEMERR;
          if (p)
            { *dlim=*p;
              *p='\0';
            }
          else *dlim='\0';
          if (isvnamech(**s))
            { (*f)->op=ID;
              i=0;
              found=0;
              while (i<nvar && !(found=(strcmp(varname[i],*s)==0))) i++;
              if (!found) return VARERR;
              else (*f)->index=i;
            }
          else 
            { (*f)->op=CONST;
              (*f)->svalue=*s;
            }
          if (p) *s=p+1;
        }
    }

  return OK;
}


char *endquote (char *p, char quote)

{ char *q,*s; 

  if (!p) return NULL;
  while (*p && *p!='\\' && *p!=quote) p++;
  if (!*p) return NULL;
  if (*p==quote) return p;
  q=p+1;
  s=q;
  if (*q=='\\' || *q=='\'' || *q=='\"')
    { do *(p++)=*q;
      while (*(q++));
    }
  return endquote(s,quote);
}


char *evalnode (struct node *p, char *x[], char *sbuff)

{ char *s,*s2,*s3,*sb,*sp,*sq,*s2p,*q1,*q2,*r;
  char wkbuff[LINE_MAX+1]; 
  char wkbuff2[LINE_MAX+1]; 
  char wkbuff3[LINE_MAX+1]; 
  int n2,n3,numerr,count,argint;
  double argdouble; 
  enum {NONE,TYPE_INT,TYPE_DOUBLE,TYPE_STRING} argtype;

  if (p==NULL) return NULL;

  switch (p->op)
    { case CONST: return p->svalue;
      case ID: return x[p->index];
      case CAT: return evalcat(p->arg,x,sbuff);
      case OR: return evalor(p->arg,x,sbuff);
      case AND: return evaland(p->arg,x,sbuff);
      default:
        { if (p->op==LENMATCH || p->op==POSOF || 
              p->op==LTRIM || p->op==RTRIM || p->op==TRIM ||
              p->op==PRINTF || 
              p->op==MIN || p->op==MAX ||
              p->op==EQ || p->op==NE ||
              p->op==LT || p->op==GT ||
              p->op==LE || p->op==GE)
            { s=evalnode(p->arg,x,wkbuff);
              s2=evalnode(p->arg2,x,wkbuff2);
            }
          else if (p->op==IF || p->op==SUBSTR || p->op==TR)
            { s=evalnode(p->arg,x,wkbuff);
              s2=evalnode(p->arg2,x,wkbuff2);
              s3=evalnode(p->arg3,x,wkbuff3);
            }
          else s=evalnode(p->arg,x,sbuff);
          switch (p->op)
            { case SUBSTR:
                if (!s || !*s) return NULL;
                if (!s2 || !*s2) n2=1;
                else 
                  { n2=sgetnum(s2,&numerr);
                    if (numerr || n2<=0) n2=1;
                  }
                if (!s3 || !*s3) n3=strlen(s);
                else
                  { n3=sgetnum(s3,&numerr);
                    if (numerr) n3=strlen(s);
                  }
                if (n2>strlen(s) || n3<=0) return NULL;
                strncpy(sbuff,s+n2-1,n3);
                if (n3<=strlen(s)) sbuff[n3]='\0';
                return sbuff;
              case MIRROR:
                if (!s || !*s) return NULL;
                sb=sbuff;
                sp=s+strlen(s)-1;
                while (sp>=s) *(sb++)=*(sp--);
                *sb='\0';
                return sbuff;                
              case LEN:
                if (!s || !*s) return "0";
                sprintf(sbuff,"%d",(int)strlen(s));
                return sbuff; 
              case LENMATCH:
                if (!s || !s2) return NULL;
                count=0;
                while (*s && *(s++)==*(s2++)) count++;
                sprintf(sbuff,"%d",count);
                return sbuff;
              case POSOF:
                if (!s || !s2 || !*s || !*s2) return NULL;
                else 
                  { s2p=strstr(s2,s); 
                    if (!s2p) return NULL; 
                    sprintf(sbuff,"%d",(int)(s2p-s2+1));
                  }
                return sbuff; 
              case TR: 
                if (!s || !*s) return NULL;
                if (!s2 || !*s2) strcpy(sbuff,s);
                else 
                  { sb=sbuff;
                    while (*s)
                      { sp=s;
                        s2p=s2; 
                        while (*sp && *s2p && *sp==*s2p) 
                          { sp++;
                            s2p++;
                          }
                        if (!*s2p)
                          { if (s3)
                              { strcpy(sb,s3);
                                sb+=strlen(s3);
                              }
                            s+=strlen(s2);
                          }
                        else 
                          { *(sb++)=*s;
                            s++;
                          }
                      }
                    *sb='\0';
                  }
                if (!*sbuff) return NULL;
                return sbuff;
              case UPPER:
                if (!s || !*s) return NULL;
                sb=sbuff;
                do
                  { if (*s>='a' && *s<='z') *(sb++)=*s-'a'+'A';
                    else *(sb++)=*s;
                  }
                while (*(++s));
                *sb='\0';
                return sbuff;
              case LOWER:
                if (!s || !*s) return NULL;
                sb=sbuff;
                do
                  { if (*s>='A' && *s<='Z') *(sb++)=*s-'A'+'a';
                    else *(sb++)=*s;
                  }
                while (*(++s));
                *sb='\0';
                return sbuff;
              case LTRIM:
                if (!s) return NULL;
                if (!s2) return s;
                count=strspn(s,s2);
                if (count<strlen(s)) strcpy(sbuff,s+count);
                else strcpy(sbuff,NODATA_STR);
                return sbuff;
              case RTRIM:
                if (!s) return NULL;
                if (!s2) return s;
                sp=s;
                count=strspn(sp,s2);
                while (count<strlen(sp)) 
                  { sp+=count;
                    sp+=strcspn(sp,s2);
                    count=strspn(sp,s2);
                  }
                count=sp-s;
                if (count!=0) 
                  { strncpy(sbuff,s,count);
                    sbuff[count]='\0';
                  }
                else strcpy(sbuff,NODATA_STR);
                return sbuff;
              case TRIM:
                if (!s) return NULL;
                if (!s2) return s;
                sq=s+strspn(s,s2);
                sp=sq;
                count=strspn(sp,s2);
                while (count<strlen(sp))
                  { sp+=count;
                    sp+=strcspn(sp,s2);
                    count=strspn(sp,s2);
                  }
                count=sp-sq;
                if (count!=0) 
                  { strncpy(sbuff,sq,count);
                    sbuff[count]='\0';
                  }
                else strcpy(sbuff,NODATA_STR);
                return sbuff;
              case PRINTF:
                if (!s || !*s || !s2) return NULL;
                argtype=NONE; 
                sp=strchr(s,'%');
                while (sp)
                  { sp++;
                    for (count=strspn(sp,"-+_#0123456789.hlL");count--;count>0)
                      { if (*sp=='_') *sp=' ';
                        sp++;
                      }
                    switch(*sp)
                      { case 'd':
                        case 'i':
                        case 'o':
                        case 'x':
                        case 'X':
                        case 'u':
                        case 'c':
                          if (argtype!=NONE) return NULL;
                          argtype=TYPE_INT;
                          argint=(int)sgetnum(s2,&numerr);
                          if (numerr) return NULL;
                          break;
                        case 'f':
                        case 'e':
                        case 'E':
                        case 'g':
                        case 'G':
                          if (argtype!=NONE) return NULL;
                          argtype=TYPE_DOUBLE;
                          argdouble=sgetnum(s2,&numerr);
                          if (numerr) return NULL;
                          break;
                        case 's':
                          if (argtype!=NONE) return NULL;
                          argtype=TYPE_STRING;
                          break;
                        case '%':
                          break;
                        default:
                          return NULL;
                          break;
                      } 
                    sp=strchr(++sp,'%');
                  }
                switch (argtype)
                  { case NONE:
                      strcpy(sbuff,s); 
                      q1=strchr(sbuff,'%');
                      while (q1)
                        { r=q1+1;
                          q2=strchr(r,'%');
                          while (*q2) *(q1++)=*(q2++);
                          *q1='\0';
                          q1=strchr(r,'%');
                        }
                      break;
                    case TYPE_INT:
                      sprintf(sbuff,s,argint);
                      break;
                    case TYPE_DOUBLE:
                      sprintf(sbuff,s,argdouble);
                      break;
                    case TYPE_STRING:
                      sprintf(sbuff,s,s2);
                      break;
                  }
                sb=sbuff;
                while (*sb)
                  { if (*sb==' ') *sb='_';
                    sb++;
                  } 
                return sbuff;
              case MIN:
                if (!s || !s2 || !*s || !*s2) return NULL;
                if (strcmp(s,s2)<=0) strcpy(sbuff,s);
                else strcpy(sbuff,s2);
                return sbuff;
              case MAX:
                if (!s && !s2) return NULL;
                if (!s2 || strcmp(s,s2)>=0) strcpy(sbuff,s);
                else strcpy(sbuff,s2);
                if (!*sbuff) return NULL;
                return sbuff;
              case IF:
                if (!s || !*s) return NULL;
                if (strcmp(s,"1")==0 || strcmp(s,"TRUE")==0) 
                  { if (!s2 || !*s2) return NULL;
                    strcpy(sbuff,s2);
                  }
                else if (strcmp(s,"0")==0 || strcmp(s,"FALSE")==0)
                  { if (!s3 || !*s3) return NULL;
                    strcpy(sbuff,s3);
                  }
                else return NULL;
                return sbuff; 
              case NOT:
                if (!s || !*s) return NULL;
                if (strcmp(s,"1")==0 || strcmp(s,"TRUE")==0) return "0";
                if (strcmp(s,"0")==0 || strcmp(s,"FALSE")==0) return "1";
                return NULL;
              case EQ:
                if ((!s || !*s) && (!s2 || !*s2)) return "1";
                if (!s || !s2 || !*s || !*s2) return "0";
                if (strcmp(s,s2)==0) return "1";
                else return "0";
              case NE:
                if ((!s || !*s) && (!s2 || !*s2)) return "0";
                if (!s || !s2 || !*s || !*s2) return "1";
                if (strcmp(s,s2)!=0) return "1";
                else return "0";
              case LT:
                if ((!s || !*s) && (!s2 || !*s2)) return "0";
                if (!s || !*s) return "1";
                if (!s2 || !*s2) return "0";
                if (strcmp(s,s2)<0) return "1";
                else return "0";
              case GT:
                if ((!s || !*s) && (!s2 || !*s2)) return "0";
                if (!s || !*s) return "0";
                if (!s2 || !*s2) return "1";
                if (strcmp(s,s2)>0) return "1";
                else return "0";
              case LE:
                if ((!s || !*s) && (!s2 || !*s2)) return "1";
                if (!s || !*s) return "1";
                if (!s2 || !*s2) return "0";
                if (strcmp(s,s2)<=0) return "1";
                else return "0";
              case GE:
                if ((!s || !*s) && (!s2 || !*s2)) return "1";
                if (!s || !*s) return "0";
                if (!s2 || !*s2) return "1";
                if (strcmp(s,s2)>=0) return "1";
                else return "0";
              default: return NULL;
            }
        }
    }
}


char *evalcat (struct node *curr, char *x[], char *sbuff)

{ char *s,*p;

  *sbuff='\0';
  p=sbuff;
  while (curr!=NULL)
    { s=evalnode(curr,x,p);
      if (s)
        { if (s!=p) strcpy(p,s);
          p+=strlen(s);
        }
      curr=curr->next;
    }
  if (strlen(sbuff)==0) return NULL;
  else return sbuff;
}


char *evalor (struct node *curr, char *x[], char *sbuff)

{ char *s;
  int missing=0;

  while (curr!=NULL)
    { s=evalnode(curr,x,sbuff);
      if (!s) missing=1;
      else if (strcmp(s,"1")==0 || strcmp(s,"TRUE")==0) return "1";
      curr=curr->next;
    }
  if (missing) return NULL;
  return "0";
}


char *evaland (struct node *curr, char *x[], char *sbuff)

{ char *s; 

  while (curr!=NULL)
    { s=evalnode(curr,x,sbuff);
      if (!s) return NULL;
      if (strcmp(s,"0")==0 || strcmp(s,"FALSE")==0) return "0";
      if (strcmp(s,"1")!=0 && strcmp(s,"TRUE")!=0) return NULL;
      curr=curr->next;
    }
  return "1";
}
