/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabtext.c - John Hemmings (Nov 2001) */

/* v1.01 - Dec 2001 - added NOTRUNC and WORD modes 
                      + field count for LINE mode */

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2 
#define NVAR 3 
#define MODE 4 

int getfieldlist_skiplf (FILE *, char *, char *[], int, int *, int *, int *);

main (int argc, char *argv[])

{ FILE *infile, *outfile;
  char varlist[LINE_MAX+1];
  char *varname[NFIELD_MAX];
  int outvar[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec d={NULL,fieldlist,field,NULL};
  char snum[11];
  enum {LINE,LINE_NOTRUNC,WORD} mode=LINE;
  int status,i,j,ilineno,nvar,noutvar,nchar,nf,lineno,firstlineno;
  char *p;

  version("1.01",VHDR,argc,argv);

  if (argc<4)
    { fprintf (stderr,"Usage: %s infile outfile nvar [notrunc|word]\n",argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (sscanf(argv[NVAR],"%d%n",&nvar,&nchar)<1 ||
      nchar!=strlen(argv[NVAR]) || nvar<0) 
    exiterror(BADEXPR,argv[0],argv[NVAR],0);
  if (nvar>NFIELD_MAX) 
    exiterror(OUTVARERR,argv[0],NULL,0);

  if (argc>4)
    { if (*argv[MODE]=='n' || *argv[MODE]=='N') mode=LINE_NOTRUNC; 
      else if (*argv[MODE]=='w' || *argv[MODE]=='W') mode=WORD;
      else exiterror(BADEXPR,argv[0],NULL,0);
    }

  if (strcmp(argv[INFILE],"-")==0) infile=stdin;
  else if ((infile=fopen(argv[INFILE],"r"))==NULL)
    exiterror(OPENINERR,argv[0],argv[INFILE],0);
  
  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else if ((outfile=fopen(argv[OUTFILE],"w"))==NULL)
    exiterror(OPENOUTERR,argv[0],argv[OUTFILE],0);

  p=varlist;
  for (i=0;i<nvar;i++)
    { sprintf(p,"f%d",i+1);
      varname[i]=p;    
      p+=strlen(p)+1;
    }
  if (mode==LINE) sprintf(p,"nf");
  else sprintf(p,"lineno");
  varname[i]=p;
  noutvar=nvar+1;

  status=writevar(outfile,0,noutvar,varname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  if (mode==WORD)
    { lineno=1; 
      while ((status=getfieldlist_skiplf(infile,d.fieldlist,d.field,nvar,
                                         &nf,&firstlineno,&lineno))==OK)
        { if (nf<nvar) for (i=nf;i<nvar;i++) d.field[i]=NULL;
          sprintf(snum,"%d",firstlineno);
          d.field[nvar]=snum;
          status=writedata(outfile,0,&d,noutvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,1);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
        }
    }
  else 
    { lineno=0;
      while ((status=getfieldlist(infile,d.fieldlist,d.field,&nf))==OK)
        { lineno++;
          for (i=0;i<nvar;i++) outvar[i]=i; 
          if (nf<nvar) for (i=nf;i<nvar;i++) d.field[i]=NULL;
          if (mode==LINE_NOTRUNC) 
            { if (nf<=nvar) ilineno=nvar;
              else ilineno=nvar*(floor((nf-1)/nvar)+1); 
              sprintf(snum,"%d",lineno);
              d.field[ilineno]=snum;
              outvar[nvar]=ilineno; 
            }
          else
            { sprintf(snum,"%d",nf); 
              d.field[nvar]=snum;
              outvar[nvar]=nvar;
            }
          status=
            writedata(outfile,0,&d,noutvar,outvar,NULL,NULL,DEFAULT,DEFAULT,1);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
          if (mode==LINE_NOTRUNC && nf>nvar)
            { j=nvar; 
              while (j<nf)
                { for (i=0;i<nvar;i++) outvar[i]=j+i; 
                  if (nf<j+nvar) for (i=nf;i<j+nvar;i++) d.field[i]=NULL; 
                  status=writedata(outfile,0,&d,noutvar,outvar,NULL,NULL,
                                   DEFAULT,DEFAULT,1);
                  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
                  j+=nvar;
                }
            }
        }
    }
  if (status!=EOF) exiterror(status,argv[0],argv[INFILE],lineno);

  fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}


int getfieldlist_skiplf (FILE *infile, char *line, char *field[], int nreq,
                         int *nf, int *firstlineno, int *lineno)

{ enum {INSPACE,INCOMMENT,INWORD} instate; 
  int c,nchar=0;
  
  *nf=0;
  instate=INSPACE;

  while ((c=fgetc(infile))!=EOF)

    { if (c=='\n') (*lineno)++;

      switch (instate)

        { case INSPACE:
            if (!strchr(WHITESPACE,c))
              { if (c=='#') instate=INCOMMENT;
                else
                  { field[*nf]=&(line[nchar]);
                    if (*nf==0) *firstlineno=*lineno;
                    instate=INWORD;
                  }
              }
            break;

          case INCOMMENT:
            if (c=='\n')
              { if ((*nf)==nreq) return OK;
                instate=INSPACE;
              }
            break;
        
          case INWORD:
            if (strchr(WHITESPACE,c) || c=='#')
               { line[nchar++]='\0';
                 (*nf)++;
                 if (c=='#') instate=INCOMMENT;
                 else 
                   { if ((*nf)==nreq) return OK; 
                     instate=INSPACE;
                   }
               }
            break;
        }

      if (instate==INWORD)
        { if (nchar==LINE_MAX) return TRUNC; 
          line[nchar++]=c;
        }
    }

  if (instate==INWORD)
    { line[nchar]='\0';
      (*nf)++;
    }

  if (*nf>0) return OK;
  return EOF;
}

