/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabview.c - John Hemmings (Mar 2005) */

/* v1.01 - Jul 2006 - changed to allow more significant digits than default for
           binary input 
   v1.02 - Nov 2008 - replaced 'readdata_special' with 'getdata' to handle 
           column data record format
   v1.03 - Feb 2010 - corrected usgae message for consistency with 'tabsqz'
*/

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2
#define LINESEP -2
#define NPAGEREC_MAX 100 
#define NPAGEREC_DEF 20 
#define PAGEWIDTH 80 

int main (int argc, char *argv[])

{ FILE *infile, *outfile;
  char *fname;
  char varlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX];
  char *outvarname[NFIELD_MAX];
  static struct {
    double x[NPAGEREC_MAX][NFIELD_MAX];
    char fieldlist[NPAGEREC_MAX][LINE_MAX+1];
    char *field[NPAGEREC_MAX][NFIELD_MAX];
  } pagebuff;
  struct datarec dpage[NPAGEREC_MAX];
  int outvpos[NFIELD_MAX];
  static double x[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec d={x,fieldlist,field,NULL};
  struct readctrlrec readctrl;
  int sd[NFIELD_MAX];
  int fw[NFIELD_MAX];
  double xval;
  int ninvar,noutvar,status,wstatus,numerr,i,j,ttb,lineno,ipage,numbered;
  int pagelen=NPAGEREC_DEF,just=DEFAULT,nspace=1,fwidthmax=LINE_MAX,
      sigdig=DEFAULT;

  version("1.03",VHDR,argc,argv);

  if (argc<2)
    { fprintf(stderr,
              "Usage: %s [=]infile [outfile [n=n_per_page] [s=col_spacing]\n",
              argv[0]);
      fprintf(stderr,
              "       [w=max_col_width] [j=justification] [d=sig_digits]]\n");
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (*argv[INFILE]=='=')
    { numbered=0;
      fname=argv[INFILE]+1;
    }
  else
    { numbered=1; 
      fname=argv[INFILE];
    }

  if (strcmp(fname,"-")==0) infile=stdin;
  else if ((infile=fopen(fname,"r"))==NULL)
    exiterror(OPENINERR,argv[0],fname,0);

  status=readvar(infile,varlist,invarname,&ninvar,&ttb,&lineno);
  if (status==OK_INLINE)
    { fclose(infile);
      infile=NULL;
    }
  else if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);

  for (i=3;i<argc;i++)
    { if (strncmp(argv[i],"n=",2)==0) 
        { if (sscanf(argv[i]+2,"%d",&pagelen)<1 || pagelen<1) 
            exiterror(BADEXPR,argv[0],argv[i],0); 
          if (pagelen>NPAGEREC_MAX) pagelen=NPAGEREC_MAX;
        }
      else if (strncmp(argv[i],"s=",2)==0) 
        { if (*(argv[i]+2)=='-') nspace=LINESEP; 
          else if (sscanf(argv[i]+2,"%d",&nspace)<1)
            exiterror(BADEXPR,argv[0],argv[i],0); 
        }
      else if (strncmp(argv[i],"w=",2)==0) 
        { if (sscanf(argv[i]+2,"%d",&fwidthmax)<1 || fwidthmax<0) 
            exiterror(BADEXPR,argv[0],argv[i],0); 
        }
      else if (strncmp(argv[i],"d=",2)==0) 
        { if (sscanf(argv[i]+2,"%d",&sigdig)<1 || sigdig<1)
            exiterror(BADEXPR,argv[0],argv[i],0);
        }
      else if (strncmp(argv[i],"j=",2)==0)
        { if (*(argv[i]+2)=='\0') exiterror(BADEXPR,argv[0],argv[i],0);
          if (strchr("lL",*(argv[i]+2))) just=LEFT; 
          else if (strchr("rR",*(argv[i]+2))) just=RIGHT; 
          else exiterror(BADEXPR,argv[0],argv[i],0); 
        }
      else exiterror(BADEXPR,argv[0],argv[i],0);
    } 

  if (numbered)
    { outvarname[0]=".";
      sd[0]=DEFAULT;
      for (i=0;i<ninvar;i++) 
        { outvpos[i]=i+1;
          outvarname[i+1]=invarname[i];
          sd[i+1]=sigdig;
        }
      noutvar=ninvar+1;
    }
  else
    { for (i=0;i<ninvar;i++) 
        { outvarname[i]=invarname[i];
          sd[i]=sigdig;
        }
      noutvar=ninvar;
    }

  if (argc<3 || strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else if ((outfile=fopen(argv[OUTFILE],"w"))==NULL)
    exiterror(OPENOUTERR,argv[0],argv[OUTFILE],0);

  for (i=0;i<pagelen;i++)
    { dpage[i].x=pagebuff.x[i];
      dpage[i].fieldlist=pagebuff.fieldlist[i];
      dpage[i].field=pagebuff.field[i];
      if (numbered) dpage[i].field[0]=NULL;
    }

  i=0;
  ipage=0;

  readctrl.nkeyvar=0;
  readctrl.keyvar=NULL;
  readctrl.nstrvar=ninvar;
  readctrl.strvar=NULL;

  do

    { status=getdata(ninvar,&sd[1],infile,ttb,&readctrl,invarname,&d,NULL,
                     &lineno);

      if (ipage==pagelen || status==EOF)

        { setcolwidths(noutvar,outvarname,ipage,dpage,sd,fwidthmax,fw);
          if (i>pagelen) fprintf(outfile,"\n");
          wstatus=writelogicalpage(outfile,noutvar,1,outvarname,fw,sd,just,
                                   nspace,PAGEWIDTH,ipage,dpage);
          if (wstatus!=OK) return wstatus;
          ipage=0;
        }

      if (status==OK)

        { j=0;
          if (numbered) 
            { dpage[ipage].x[j++]=(double)(i+1);
              overlaydata(&dpage[ipage],&d,ninvar,outvpos,NULL,
                          dpage[ipage].fieldlist);
            }
          else copydata(&dpage[ipage],&d,ninvar,NULL);
          ipage++; 
        }

      i++;
    }

  while (status==OK);
  if (status!=EOF) exiterror(status,argv[0],argv[INFILE],lineno); 
  
  if (infile) fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}

