/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* toolio.c - Table Toolbox library - John Hemmings (Oct 2000) */

/* v1.01 - Nov 2000 - new function 'outfopen' 
   v1.02 - Jan 2001 - new functions 'validvname' and 'isvnamech' 
           and new error type 'VFMTERR' +
           variable name validation added to 'readvar' function +
           'about' banner changed
   v1.03 - Nov 2001 - added missing filename argument for 'OUTERR' error msg 
   v1.04 - Apr 2002 - TAB options added to 'putcolsep' function +
           new function 'writefdata' +
           'writevar' and 'putcolsep' changed to handle new col sep flags + 
           replaced automatic storage with static for speed + 
           new error type 'OUTTRUNC' 
   v1.05 - Jan 2003 - Allow '\' in variable names
   v1.06 - Sep 2003 - 'about' banner changed 
   v1.07 - Jan 2004 - 'readdata' and 'mreaddata' changed to properly 
           handle key variables for which string format is requested when the 
           input data is not binary +
           changed 'VARERR' message to output filename too if supplied +
           added new error type 'BADDATA' +
           allow 'sputnum' call with NULL output buffer for just finding 
           length of output string + 
           replaced automatic storage with static in 'outfopen' and 
           'ttbname'
   v1.08 - Mar 2005 - new functions 'writelogicalpage', 'setcolwidths',  
           'setwidth' and 'overlaydata' 
   v1.09 - May 2006 - 'outfopen' changed to include stdout recognition +
           changed 'about' banner
   v1.10 - Jul 2006 - new functions 'readdata_special' and 'mreaddata_special'
   v1.11 - Sep 2006 - 'strconvback' buffer now terminated correctly with null
           to avoid potential overrun 
   v1.12 - Oct 2006 - fixed 'readdata_special' and 'mreaddata_special' to avoid
           conversion of text fields when the no. of significant digits is
           not given explicitly
   v1.13 - Aug 2008 - changed error message from "error in" to "error running"
   v1.14 - Nov 2008 - modified 'readvar' to read a data record in column fmt +
           added associated status 'OK_INLINE' to 'exiterror' +
           new function 'getdata' +
           added quotes around file names in error messages + 
           implemented 'COLSEP_LINE' in 'writelogicalpage' +
           fixed bug in 'basewidth' calculation for repeated variables 
   v1.15 - Feb 2009 - new function 'roundval' - used for explicit rounding in 
           'putnum' and 'sputnum' prior to calling 'fprintf'/'sprintf' to 
           avoid ambiguity due to different C library implementations 
           ("round half away from zero" convention NOT "round half to even"),
           used also in keyconv in place of 'sprintf' +
           'about' banner changed +
           changed input routines 'readdata_special', 'mreaddata_special and 
           'getdata' to properly handle conversions:
             - use 'strconvback' to handle all sig. digit conversions not
               just key conversions (fixes problem due to string length 
               extension)
             - only variables to be processed as strings are subject to
               explicit sig. digit conversion on input (others are converted to
               numeric values to force processing on output)
             - ensure keys always have default sig. digits on input 
   v1.16 - May 2009 - fixed bug in 'roundval' for requested precision sigdig>9 
           (also trapped x=0 to avoid potential problems with infinite values) 
   v1.17 - Mar 2010 - fixed accounting bug in 'strconvback' that stopped 
           trapping of buffer overrun from working +
           corrected error message for status OK_INLINE ("column data record 
           format" is now referred to as "transposed table format") 
   v1.18 - Mar 2012 - fixed 'roundval' to avoid overflow for very small numbers
   v1.19 - Nov 2013 - 'about' banner changed
*/ 

#define VLIB "1.19"
#include "toolbox.h"

int strconvback (struct datarec *, int, int, int [], int [], int *);

void version (char *vsrc, char *vhdrsrc, int argc, char *argv[])

{ int about=0;

  if (argc!=2) return;  
  
  if (strcmp(argv[1],"about?")==0)
    { about=1;
      fprintf(stderr,"------------------------------------------------\n");
      fprintf(stderr,"                                                \n");
      fprintf(stderr,"               - TABLE TOOLBOX -                \n");
      fprintf(stderr,"                                                \n");
      fprintf(stderr,"             Author: John Hemmings              \n");
      fprintf(stderr,"        UK National Oceanography Centre         \n");
      fprintf(stderr," Copyright Natural Environment Research Council \n");
      fprintf(stderr,"                Licence: CeCILL                 \n");
      fprintf(stderr,"                                                \n");
      fprintf(stderr,"       Last core library update: Nov 2013       \n");
      fprintf(stderr,"                                                \n");
      fprintf(stderr,"------------------------------------------------\n");
    }

  if (about || strcmp(argv[1],"version?")==0)
    { fprintf(stderr,
              "program '%s' version %s (header %s)\n",argv[0],vsrc,vhdrsrc);
      fprintf(stderr,"library version %s (header %s)\n",VLIB,VHDR);
      exit(EXIT_FAILURE);
    }
}


int readvar (FILE *infile, char *varlist, char *varname[], int *nvar, int *ttb,
             int *gcount)

{ static char flistbuff[LINE_MAX+1];
  static char *fbuff[NFIELD_MAX];
  static int inlinefmt,status,i,nf;
  static char *s;

  inlinefmt=0;
  *ttb=0;
  *gcount=1;
  while ((status=getfieldlist(infile,varlist,varname,nvar))==OK && *nvar==0)
    (*gcount)++;
  if (status!=OK) return status;

  if (strcmp(varname[0],"!TTB")==0)
    { *ttb=1;
      if (*nvar==1)
        { while ((status=getfieldlist(infile,varlist,varname,nvar))==OK &&
                 *nvar==0) (*gcount)++;
          if (status!=OK) return status;
        }
      else
        { for (i=0;i<(*nvar)-1;i++) varname[i]=varname[i+1];
          (*nvar)--;
        }
    }

  else 

    { if (*nvar==2 && strcmp(varname[0],"NAME")==0 && 
          strcmp(varname[1],"VALUE")==0)
        { inlinefmt=1;
          (*nvar)=0;
          s=varlist;
          while ((status=getfieldlist(infile,flistbuff,fbuff,&nf))==OK)
            { (*gcount)++;
              if (nf>0)
                { if (nf!=2) return FMTERR;
                  if ((*nvar)>NFIELD_MAX) return FLDTRUNC;
                  if (s-varlist+strlen(fbuff[0])+1+strlen(fbuff[1])>LINE_MAX) 
                    return TRUNC;
                  varname[*nvar]=s;
                  strcpy(s,fbuff[0]);
                  s+=strlen(s)+1;
                  strcpy(s,fbuff[1]);
                  s+=strlen(s)+1;
                  (*nvar)++;
	        }
            }
          if (status!=EOF) return status;
          *gcount=0;
        }

      else if (*nvar==1 && strcmp(varname[0],"NAME")==0)
        { inlinefmt=1;
          (*nvar)=0;
          s=varlist;
          while ((status=getfieldlist(infile,flistbuff,fbuff,&nf))==OK)
            { (*gcount)++;
              if (nf>0)
                { if (nf!=1) return FMTERR;
                  if ((*nvar)>NFIELD_MAX) return FLDTRUNC;
                  if (s-varlist+strlen(fbuff[0])>LINE_MAX) return TRUNC;
                  varname[*nvar]=s;
                  strcpy(s,fbuff[0]);
                  s+=strlen(s)+1;
                  (*nvar)++;
                }
            }
          if (status!=EOF) return status;
          *gcount=-1;
        }
    }

  for (i=0;i<(*nvar);i++)
    if (!validvname(varname[i])) return VFMTERR;

  if (inlinefmt) return OK_INLINE;
  return OK;
}


int validvname (char *p)

{ if (!isvnamech(*p)) return 0;
  while (*(++p))
    if (!isvnamech(*p) && !isdigit(*p) && *p!='.') return 0;

  return 1;
}


int isvnamech (int c)

{ if (isalpha(c)) return 1;
  if (c=='$') return 1;
  if (c=='%') return 1;
  if (c=='^') return 1;
  if (c=='_') return 1;
  if (c==':') return 1;
  if (c=='~') return 1;
  if (c=='\\') return 1;
  return 0;
}


int writevar (FILE *outfile, int ttb, int nvar, char *varname[], int fwidth[],
              int nspace)

{ static int fw,i,nchar;

  if (ttb)
    if (fprintf(outfile,"!TTB ")<0) return OUTERR;

  if (nspace==DEFAULT)
    { if (ttb) nspace=1;
      else nspace=NSPACE_DEF;
    }

  if (!ttb && nspace==COLSEP_LINE)
    { if (putrowsep(outfile,nvar,fwidth,nspace)<0) return OUTERR;
    }

  for (i=0;i<nvar;i++)

    { if (fwidth==NULL || fwidth[i]==DEFAULT)
        { if (ttb) fw=0;
          else fw=FWIDTH_DEF;
        }
      else fw=fwidth[i];

      nchar=fprintf(outfile,"%-*s",fw,varname[i]);
      if (nchar<=0) return OUTERR;
      if (i<nvar-1)
        if (putcolsep(outfile,nspace,nchar-fw)<0) return OUTERR;
    }

  if (fprintf(outfile,"\n")<0) return OUTERR;
  if (!ttb)
    { if (nspace==COLSEP_LINE)
        { if (putrowsep(outfile,nvar,fwidth,nspace)<0) return OUTERR;
        }
      else if (fprintf(outfile,"\n")<0) return OUTERR;
    }

  return OK;
}


int indexlist (int nreq, char *req[], int nf, char *field[], int index[])

{ static int i,j,found,count;

  count=0;
  for (i=0;i<nreq;i++)
    { j=0;
      found=0;
      while (j<nf && !(found=(strcmp(field[j],req[i])==0)))
        j++;
      if (found)
        { index[i]=j;
          count++;
        }
      else index[i]=-1;
    }

  return count;
}


int readdata_special (FILE *infile, int ttb, int nvar, int nkeyvar, 
                      int keyvar[], int nstrvar, int strvar[], int sigdig[], 
                      struct datarec *d, int *gcount)

{ static int maxchar;
  static char *s;
  static int status,len,nf,i,istr,numerr;

  if (sigdig && nstrvar>0)
    for (i=0;i<nkeyvar;i++) sigdig[i]=DEFAULT;

  if (ttb)

    { if (*gcount>0) *gcount=0;
      (*gcount)--;

      status=getbindata(infile,nvar,d->x);
      if (status!=OK) return status;

      if (d->field)
        for (i=0;i<nvar;i++) d->field[i]=NULL;
      if (nkeyvar>0) keyconv(d,nkeyvar,keyvar);
      if (nstrvar==0) return OK; 
 
      maxchar=LINE_MAX;
      s=d->fieldlist;
      for (i=0;i<nstrvar;i++)
        { if (strvar) istr=strvar[i];
          else istr=i;
          if (sigdig)
            len=sputnum(s,maxchar,d->x[istr],0,sigdig[istr]);
          else
            len=sputnum(s,maxchar,d->x[istr],0,DEFAULT);
          if (len<0) return TRUNC;
          d->field[istr]=s;
          s+=len+1;
          maxchar-=len+1;
          d->x[istr]=NODATA;
        }
      return OK; 
    }

  else

    { (*gcount)++;

      while ((status=getfieldlist(infile,d->fieldlist,d->field,&nf))==OK
             && nf==0) (*gcount)++;
      if (status!=OK) return status;
      if (nf!=nvar) return FMTERR;
      if (!d->x) return OK;

      for (i=0;i<nvar;i++) d->x[i]=NODATA;
      if (sigdig)
        { for (i=0;i<nvar;i++) if (sigdig[i]!=DEFAULT) 
            { d->x[i]=sgetnum(d->field[i],&numerr);
              if (!numerr) d->field[i]=NULL;
            }
        }
      if (nkeyvar>0) keyconv(d,nkeyvar,keyvar);
      if (nstrvar==0) return OK; 

      status=strconvback(d,nvar,nstrvar,strvar,sigdig,&nf);
      if (status!=OK || nf==0) return status;

      s=d->fieldlist;
      for (i=0;i<nvar;i++) if (d->field[i])
        { strcpy(s,d->field[i]);
          d->field[i]=s;
          s+=strlen(s)+1;
        } 
      
      return OK;
    }
}


int readdata (FILE *infile, int ttb, int nvar, int nkeyvar, int keyvar[],
              int nstrvar, int strvar[], struct datarec *d, int *gcount)

{ return readdata_special 
           (infile,ttb,nvar,nkeyvar,keyvar,nstrvar,strvar,NULL,d,gcount);
}


int mreaddata_special (FILE *infile, int ttb, int nvar, int nkeyvar, 
                       int keyvar[], int nstrvar, int strvar[], int sigdig[], 
                       struct datarec **d, int *gcount)

{ static char flistbuff[LINE_MAX+1],*s;
  static double xbuff[NFIELD_MAX];
  static int maxchar;
  static int status,len,nf,i,istr,numerr;
  
  *d=(struct datarec *)malloc((size_t)sizeof(struct datarec));
  if (*d==NULL) return MEMERR;
  if (nstrvar==nvar) (*d)->x=xbuff;
  else
    { (*d)->x=(double *)malloc((size_t)nvar*sizeof(double));
      if ((*d)->x==NULL) return MEMERR;
    }
  if (ttb && nstrvar==0)
    { (*d)->field=NULL;
      (*d)->fieldlist=NULL;
    }
  else
    { (*d)->field=(char **)malloc((size_t)nvar*sizeof(char *));
      if ((*d)->field==NULL) return MEMERR;
    }
  (*d)->next=NULL;

  if (sigdig && nstrvar>0)
    for (i=0;i<nkeyvar;i++) sigdig[i]=DEFAULT;

  if (ttb)

    { if (*gcount>0) *gcount=0;
      (*gcount)--;

      status=getbindata(infile,nvar,(*d)->x);
      if (status!=OK) return status;
      if (nkeyvar>0) keyconv(*d,nkeyvar,keyvar);
      if (nstrvar==0) return OK;

      for (i=0;i<nvar;i++) (*d)->field[i]=NULL;
      maxchar=LINE_MAX;
      s=flistbuff;
      for (i=0;i<nstrvar;i++)
        { if (strvar) istr=strvar[i];
          else istr=i;
          if (sigdig)
            len=sputnum(s,maxchar,(*d)->x[istr],0,sigdig[istr]);
          else
            len=sputnum(s,maxchar,(*d)->x[istr],0,DEFAULT);
          if (len<0) return TRUNC;
          (*d)->field[istr]=s;
          s+=len+1;
          maxchar-=len+1;
          (*d)->x[istr]=NODATA;
        }

      (*d)->fieldlist=(char *)malloc((size_t)(s-flistbuff));
      if ((*d)->fieldlist==NULL) return MEMERR;
      s=(*d)->fieldlist;
      for (i=0;i<nstrvar;i++)
        { if (strvar) istr=strvar[i];
          else istr=i;
          strcpy(s,(*d)->field[istr]);
          (*d)->field[istr]=s;
          s+=strlen(s)+1;
        }
    }

  else

    { (*gcount)++;

      while ((status=getfieldlist(infile,flistbuff,(*d)->field,&nf))==OK
             && nf==0) (*gcount)++;
      if (status!=OK) return status;
      if (nf!=nvar) return FMTERR;

      for (i=0;i<nvar;i++) (*d)->x[i]=NODATA;
      if (sigdig)
        { for (i=0;i<nvar;i++) if (sigdig[i]!=DEFAULT)
            { (*d)->x[i]=sgetnum((*d)->field[i],&numerr);
              if (!numerr) (*d)->field[i]=NULL;
            }
        }
      if (nkeyvar>0) keyconv(*d,nkeyvar,keyvar);
      if (nstrvar>0) 
        { status=strconvback(*d,nvar,nstrvar,strvar,sigdig,&nf);
          if (status!=OK) return status;
        }

      len=0;
      for (i=0;i<nvar;i++) 
        if ((*d)->field[i]) len+=strlen((*d)->field[i])+1;
      (*d)->fieldlist=(char *)malloc((size_t)len);
      if ((*d)->fieldlist==NULL) return MEMERR;
      s=(*d)->fieldlist;
      for (i=0;i<nvar;i++) if ((*d)->field[i])
        { strcpy(s,(*d)->field[i]);
          (*d)->field[i]=s;
          s+=strlen(s)+1;
        }
    }

  if (nstrvar==nvar) (*d)->x=NULL; 
  return OK;
}


int mreaddata (FILE *infile, int ttb, int nvar, int nkeyvar, int keyvar[],
               int nstrvar, int strvar[], struct datarec **d, int *gcount)

{ return mreaddata_special 
           (infile,ttb,nvar,nkeyvar,keyvar,nstrvar,strvar,NULL,d,gcount);
}          


int getdata (int nvar, int sigdig[], FILE *infile, int ttb, 
             struct readctrlrec *readctrl, char *inlinevar[], struct datarec *d,
             struct datarec **dp, int *gcount)

{ static struct readctrlrec readctrl_default={0,NULL,0,NULL}; 
  static struct readctrlrec *rc;
  static char *s;
  static int status,numerr,i,len,n;

  if (readctrl) rc=readctrl;
  else rc=&readctrl_default;

  if (!infile)

    { if (*gcount<0) return EOF;

      if (d && d->x)
        for (i=0;i<nvar;i++) d->x[i]=NODATA;
  
      else if (!d)
        { *dp=(struct datarec *)malloc((size_t)sizeof(struct datarec));
          if (*dp==NULL) return MEMERR;
          (*dp)->x=(double *)malloc((size_t)nvar*sizeof(double));
          if ((*dp)->x==NULL) return MEMERR;
          for (i=0;i<nvar;i++) (*dp)->x[i]=NODATA;
          (*dp)->field=(char **)malloc((size_t)nvar*sizeof(char *));
          if ((*dp)->field==NULL) return MEMERR;
          (*dp)->next=NULL;
        }

      if (sigdig && rc->nstrvar>0)
        for (i=0;i<rc->nkeyvar;i++) sigdig[i]=DEFAULT;

      for (i=0;i<nvar;i++)
        { s=inlinevar[i]; 
          s+=strlen(s)+1;
          if (sigdig && sigdig[i]!=DEFAULT)
            { if (d) d->x[i]=sgetnum(s,&numerr);
              else (*dp)->x[i]=sgetnum(s,&numerr);
              if (!numerr) s=NULL;
            }
          if (d) d->field[i]=s;
          else (*dp)->field[i]=s;
        }

      if (d)

        { if (rc->nkeyvar>0) keyconv(d,rc->nkeyvar,rc->keyvar);
          if (rc->nstrvar>0) 
            { status=strconvback(d,nvar,rc->nstrvar,rc->strvar,sigdig,&n);
              if (status!=OK) return status;
              if (n>0) 
                { s=d->fieldlist;
                  for (i=0;i<nvar;i++) if (d->field[i])
                    { strcpy(s,d->field[i]);
                      d->field[i]=s;
                      s+=strlen(s)+1;
                    }
                }
            }
        }

      else

        { if (rc->nkeyvar>0) keyconv(*dp,rc->nkeyvar,rc->keyvar);
          if (rc->nstrvar>0)  
            { status=strconvback(*dp,nvar,rc->nstrvar,rc->strvar,sigdig,&n);
              if (status!=OK) return status;
            }   

          len=0;
          for (i=0;i<nvar;i++) 
            if ((*dp)->field[i]) len+=strlen((*dp)->field[i])+1;

          (*dp)->fieldlist=(char *)malloc((size_t)len);
          if ((*dp)->fieldlist==NULL) return MEMERR;
          s=(*dp)->fieldlist;
          for (i=0;i<nvar;i++) 
            { strcpy(s,(*dp)->field[i]);
              (*dp)->field[i]=s;
              s+=strlen(s)+1;
            }
        }

      *gcount=-1;
      return OK;
    } 

  if (d)
    return readdata_special
             (infile,ttb,nvar,rc->nkeyvar,rc->keyvar,rc->nstrvar,rc->strvar,
              sigdig,d,gcount);
  else
    return mreaddata_special 
             (infile,ttb,nvar,rc->nkeyvar,rc->keyvar,rc->nstrvar,rc->strvar,
              sigdig,dp,gcount);
}


void keyconv (struct datarec *d, int nkeyvar, int keyvar[])

{ static char sbuff[LINE_MAX+1];
  static char **f;
  static double *x;
  static int ikey,i,numerr;

  f=NULL;
  for (i=0;i<nkeyvar;i++)
    { if (keyvar) ikey=keyvar[i];
      else ikey=i;
      x=&d->x[ikey];
      if (*x==NODATA && d->field)
        { f=&d->field[ikey];
          *x=sgetnum(*f,&numerr);
        }
      if (*x!=NODATA)
        { *x=roundval(*x,SIGDIG_DEF);
          if (f) *f=NULL;
        }
    }
}


int strconvback (struct datarec *d, int nvar, int nstrvar, int strvar[], 
                 int sigdig[], int *nstrconv)

/* N.B. When fields are converted to double in 'keyconv' and converted back in 
   this function it IS possible for them to be longer than in the input stream,
   so truncation at end of buffer is allowed for */ 

{ static char sbuff[LINE_MAX+1],*s;
  static int strconvflag[NFIELD_MAX];
  static int maxchar,len,i,j;

  *nstrconv=0;
  if (strvar)
    { for (i=0;i<nvar;i++)
        { if (!d->field[i])
            { strconvflag[i]=0;
              for (j=0; j<nstrvar && !strconvflag[i] ;j++) if (strvar[j]==i)
                { (*nstrconv)++;
                  strconvflag[i]=1;
                }
            }
        }    
    }
  else 
    { for (i=0;i<nstrvar;i++)
        { if (!d->field[i])
            { (*nstrconv)++;
              strconvflag[i]=1;
            }
        }
      for (i=nstrvar;i<nvar;i++) strconvflag[i]=0;
    }

  if (*nstrconv==0) return OK;

  sbuff[LINE_MAX]='\0';
  maxchar=LINE_MAX;
  s=sbuff;
  for (i=0;i<nvar;i++)
    { if (d->field[i])
        { strncpy(s,d->field[i],maxchar);
          len=strlen(s);
          if (len<strlen(d->field[i])) return TRUNC;
          d->field[i]=s;
          s+=len+1;
          maxchar-=len+1;
        }
      else if (strconvflag[i])
        { if (sigdig) len=sputnum(s,maxchar,d->x[i],0,sigdig[i]);
          else len=sputnum(s,maxchar,d->x[i],0,DEFAULT);
          if (len<0) return TRUNC;
          d->field[i]=s;
          s+=len+1;
          maxchar-=len+1;
          d->x[i]=NODATA;
        }
    }

  return OK;
}


void copydata (struct datarec *d1, struct datarec *d2, int nvar, int v[])

{ static char *s;
  static int i;

  if (d1->x!=NULL)
    for (i=0;i<nvar;i++)
      { if (d2->x==NULL) d1->x[i]=NODATA;
        else
          { if (v) d1->x[i]=d2->x[v[i]];
            else d1->x[i]=d2->x[i];
          }
      }

  if (d1->field!=NULL)
    { s=d1->fieldlist;
      for (i=0;i<nvar;i++)
        { if (d2->field==NULL ||
              (v && d2->field[v[i]]==NULL) ||
              (v==NULL && d2->field[i]==NULL)) d1->field[i]=NULL;
          else
            { if (v) strcpy(s,d2->field[v[i]]);
              else strcpy(s,d2->field[i]);
              d1->field[i]=s;
              s+=strlen(s)+1;
            }
        }
    }
}


void overlaydata (struct datarec *d1, struct datarec *d2, int nvar, 
                  int v1[], int v2[], char *s)

{ static int i;

  if (d1->x!=NULL)
    for (i=0;i<nvar;i++)
      { if (d2->x==NULL) 
          { if (v1) d1->x[v1[i]]=NODATA;
            else d1->x[i]=NODATA;
          }
        else
          { if (v2) 
              { if (v1) d1->x[v1[i]]=d2->x[v2[i]];
                else d1->x[i]=d2->x[v2[i]];
              }
            else 
              { if (v1) d1->x[v1[i]]=d2->x[i];
                else d1->x[i]=d2->x[i];
              }
          }
      }

  if (d1->field!=NULL)
    { for (i=0;i<nvar;i++)
        { if (d2->field==NULL ||
              (v2 && d2->field[v2[i]]==NULL) ||
              (v2==NULL && d2->field[i]==NULL)) 
            { if (v1) d1->field[v1[i]]=NULL;
              else d1->field[i]=NULL;
            }
          else
            { if (v2) strcpy(s,d2->field[v2[i]]);
              else strcpy(s,d2->field[i]);
              if (v1) d1->field[v1[i]]=s; 
              else d1->field[i]=s;
              s+=strlen(s)+1;
            }
        }
    }
}


int writedata (FILE *outfile, int ttb, struct datarec *d, int nvar, int v[],
               int fwidth[], int sigdig[], int just, int nspace, int term)

{ static double *x,xtmp;
  static char *s;
  static int status,fw,sig,nchar,i;

  if (nspace==DEFAULT) nspace=NSPACE_DEF;

  for (i=0;i<nvar;i++)

    { if (d && d->field)
        { if (v) s=d->field[v[i]];
          else s=d->field[i];
        }
      else s=NULL;

      if (d && d->x)
        { if (v) x=&d->x[v[i]];
          else x=&d->x[i];
        }
      else
        { xtmp=NODATA;
          x=&xtmp;
        }

      if (ttb)
        { if (*x==NODATA) *x=sgetnum(s,&status);
          status=putbindata(outfile,1,x);
          if (status!=OK) return status;
        }
      else

        { if (fwidth) fw=fwidth[i];
          else fw=DEFAULT;

          if (s) nchar=putstr(outfile,s,fw,just);
          else
            { if (sigdig) sig=sigdig[i];
              else sig=DEFAULT;
              nchar=putnum(outfile,*x,fw,sig);
            }
          if (nchar<=0) return OUTERR;
          if (i<nvar-1 || !term)
            if (putcolsep(outfile,nspace,nchar-fw)<0) return OUTERR;
        }
    }

  if (term && !ttb)
    if (fprintf(outfile,"\n")<0) return OUTERR;
  return OK;
}


int writefdata (FILE *outfile, struct datarec *d, int nvar, int v[],
                char *fmt[], int sigdig[], int nspace, int term)

{ static double *x,xtmp;
  static char *s,sbuff[LINE_MAX+1];
  static int nchar,i,sd;

  if (nspace==DEFAULT) nspace=NSPACE_DEF;

  for (i=0;i<nvar;i++)

    { if (d && d->field)
        { if (v) s=d->field[v[i]];
          else s=d->field[i];
        }
      else s=NULL;

      if (d && d->x)
        { if (v) x=&d->x[v[i]];
          else x=&d->x[i];
        }
      else
        { xtmp=NODATA;
          x=&xtmp;
        }

      if (s) nchar=fprintf(outfile,fmt[i],s);
      else 
        { if (sigdig) sd=sigdig[i];
          else sd=SIGDIG_DEF;
          sputnum(sbuff,LINE_MAX,*x,0,sd);
          nchar=fprintf(outfile,fmt[i],sbuff);
        }
      if (nchar<=0) return OUTERR;
      if (i<nvar-1 || !term)
        if (putcolsep(outfile,nspace,0)<0) return OUTERR;
    }

  if (term)
    if (fprintf(outfile,"\n")<0) return OUTERR;

  return OK;
}


int writelogicalpage (FILE *outfile, int nvar, int nrepeatvar, char *varname[],
                      int fw[], int sd[], int just, int nspace, int pagewidth, 
                      int nrec, struct datarec d[])

{ static char *pagevarname[NFIELD_MAX];
  static int pagevar[NFIELD_MAX];
  static int pagefw[NFIELD_MAX];
  static int pagesd[NFIELD_MAX];
  static int basewidth,pwidth,npagevar,status,i,j;

  if (!outfile) return OK;
  if (nvar>0 && !varname) return OK;

  if (nspace==DEFAULT || (nspace<0 && nspace!=COLSEP_LINE)) nspace=NSPACE_DEF;

  if (nvar<=nrepeatvar)

    { status=writevar(outfile,0,nvar,varname,fw,nspace);
      if (status!=OK) return status;

      if (d) for (i=0;i<nrec;i++)
        { status=
            writedata(outfile,0,&d[i],nvar,NULL,fw,sd,just,nspace,1);
          if (status!=OK) return status;
        }

      return OK;
    }

  /* Write repeating variables plus one or more additional variables
     on each physical page, up to the width limit 'pagewidth' */

  basewidth=0;
  for (i=0;i<nrepeatvar;i++)
    { pagevarname[i]=varname[i];
      pagevar[i]=i;
      pagefw[i]=fw[i];
      if (sd) pagesd[i]=sd[i];
      else pagesd[i]=DEFAULT;
      if (nspace==COLSEP_LINE) basewidth+=(3+fw[i]);
      else basewidth+=(nspace+fw[i]);
    }

  while (i<nvar)

    { if (i!=nrepeatvar) fprintf(outfile,"\n");

      pwidth=basewidth+fw[i];
      npagevar=nrepeatvar;
      do
        { pagevarname[npagevar]=varname[i];
          pagevar[npagevar]=i;
          pagefw[npagevar]=fw[i];
          if (sd) pagesd[npagevar]=sd[i];
          else pagesd[npagevar]=DEFAULT;
          npagevar++;
          i++;
        }
      while (i<nvar && 
             (pwidth+=((nspace==COLSEP_LINE)?3:nspace)+fw[i])<pagewidth);

      status=writevar(outfile,0,npagevar,pagevarname,pagefw,nspace);
      if (status!=OK) return status;

      for (j=0;j<nrec;j++)
        { status=writedata(outfile,0,&d[j],npagevar,pagevar,pagefw,pagesd,
                           just,nspace,1);
          if (status!=OK) return status;
        }
    }

  return OK;
}


void setcolwidths (int nvar, char *varname[], int nrec, struct datarec d[], 
                   int sd[], int cwidthmax, int fw[])

{ static int sigdig,i,j;

  for (i=0;i<nvar;i++) fw[i]=strlen(varname[i]);

  for (i=0;i<nrec;i++)
    { for (j=0;j<nvar;j++)
        { if (d[i].field[j])
            fw[j]=setwidth(fw[j],strlen(d[i].field[j]),cwidthmax);
          else
            { if (sd) sigdig=sd[j];
              else sigdig=DEFAULT;
              fw[j]=setwidth(fw[j],sputnum(NULL,cwidthmax,d[i].x[j],0,sigdig),
                             cwidthmax);
            }
        }
    }
}


int setwidth (int currwidth, int fieldlen, int limit)

{ int w;

  if (fieldlen>currwidth) w=fieldlen;
  else w=currwidth;
  if (w>limit) w=limit;

  return w;
}


int datacomp (int nvar, struct datarec *d1, int v1[], struct datarec *d2,
              int v2[], int cmptype)

/* Numeric comparison: numeric representation is used if available, else
   strings are converted to numbers if possible (Conversion results are
   saved in data record if storage space is available)
   - Non-numeric strings precede numbers (in order given by string comparison)
   - NODATA values precede valid numbers
   String comparison: any numeric representation is ignored
   - NODATA_STR values precede valid strings
   - NULL strings precede NODATA_STR */

{ static char *s1,*s2;
  static double *x1,*x2,x1tmp,x2tmp;
  static int numerr1,numerr2,i,c;

  for (i=0;i<nvar;i++)

    { if (d1->field)
        { if (v1) s1=d1->field[v1[i]];
          else s1=d1->field[i];
        }
      else s1=NULL;

      if (d2->field)
        { if (v2) s2=d2->field[v2[i]];
          else s2=d2->field[i];
        }
      else s2=NULL;

      if (cmptype==NUMCOMP)

        { if (d1->x)
            { if (v1) x1=&d1->x[v1[i]];
              else x1=&d1->x[i];
            }
          else
            { x1tmp=NODATA;
              x1=&x1tmp;
            }

          if (d2->x)
            { if (v2) x2=&d2->x[v2[i]];
              else x2=&d2->x[i];
            }
          else
            { x2tmp=NODATA;
              x2=&x2tmp;
            }

          numerr1=0;
          numerr2=0;
          if (*x1==NODATA) *x1=sgetnum(s1,&numerr1);
          if (*x2==NODATA) *x2=sgetnum(s2,&numerr2);
          if (numerr1)
            { if (!numerr2) return -1;
            }
          else if (numerr2) return 1;
          else if (*x1==NODATA && *x2!=NODATA) return -1;
          else if (*x2==NODATA && *x1!=NODATA) return 1;
          else
            { if (*x1<*x2) return -1;
              if (*x1>*x2) return 1;
            }
        }

      if (cmptype==STRCOMP || (cmptype==NUMCOMP && numerr1 && numerr2))
        { if (s1==NULL)
            { if (s2==NULL) c=0;
              else return -1;
            }
          else if (s2==NULL) return 1;
          else if (strcmp(s1,NODATA_STR)==0)
            { if (strcmp(s2,NODATA_STR)==0) c=0;
              else return -1;
            }
          else if (strcmp(s2,NODATA_STR)==0) return 1;
          else c=strcmp(s1,s2);
          if (c!=0) return c;
        }
    }
  return 0;
}


int sortcomp (struct datarec *d1, struct datarec *d2, struct sortrec *spec)

{ static int i,c;

  for (i=0;i<spec->nvar;i++)
    { c=spec->sense[i]*
        datacomp(1,d1,&spec->var[i],d2,&spec->var[i],spec->ctype[i]);
      if (c!=0) return c;
    }

  return c;
}


int getfieldlist (FILE *infile, char *line, char *field[], int *nf)

{ static char *s;
  static int trunc,len,c;

  trunc=0;
  *nf=0;
  if (fgets(line,LINE_MAX+2,infile)==NULL) return EOF;

  len=strlen(line);
  if (len>LINE_MAX && line[LINE_MAX]!='\n')
    { trunc=1;
      while ((c=fgetc(infile))!='\n' && c!=EOF);
    }

  if (line[0]=='#') return OK;
  strtok(line,"#");
  if (trunc && strlen(line)==len) return TRUNC;

  s=strtok(line,WHITESPACE);
  while (s && *nf<NFIELD_MAX)
    { field[(*nf)++]=s;
      s=strtok(NULL,WHITESPACE);
    }
  if (s) return FLDTRUNC;
  return OK;
}


int putstr (FILE *outfile, char *s, int fwidth, int just)

{ static int numerror;

  if (fwidth==DEFAULT) fwidth=FWIDTH_DEF;
  if (just==DEFAULT)
    { sgetnum(s,&numerror);
      if (numerror) just=LEFT;
      else just=RIGHT;
    }
  if (just==RIGHT)
    return fprintf(outfile,"%*s",fwidth,s);
  else
    return fprintf(outfile,"%-*s",fwidth,s);
}


double roundval (double x, int sigdig)

{ static double dp,scale;

  if (x==0) 
    { x=0; /* change -0 to 0 */
      return x;
    }

  if (sigdig<1) sigdig=SIGDIG_DEF;
  else if (sigdig>DBL_DIG-1) sigdig=DBL_DIG-1;
  dp=sigdig-ceil(log10(fabs(x)));
  if (dp<POW10_EXP_LIM)
    { scale=pow(10,dp);
      if (x<0) x=ceil(scale*x-0.5)/scale;
      else x=floor(scale*x+0.5)/scale;
      if (x==0) x=0; /* change -0 to 0 */
    }
  return x;
}


int putnum (FILE *outfile, double x, int fwidth, int sigdig)

{ if (fwidth==DEFAULT) fwidth=FWIDTH_DEF;
  if (x==NODATA) return fprintf(outfile,"%-*s",fwidth,NODATA_STR);
  if (sigdig==DEFAULT || sigdig<1) sigdig=SIGDIG_DEF;
  x=roundval(x,sigdig); 
  return fprintf(outfile,"%*.*g",fwidth,sigdig,x);
}


int putrowsep (FILE *outfile, int nvar, int fwidth[], int nspace)

{ static int len,i;

  len=0;
  if (nspace<0) nspace=3;
  for (i=0;i<nvar-1;i++) len+=fwidth[i]+nspace;
  len+=fwidth[nvar-1];
  for (i=0;i<len;i++)
    if (fputc('-',outfile)==EOF) return -1;
  if (fputc('\n',outfile)==EOF) return OUTERR;
  return len+1;
}


int putcolsep (FILE *outfile, int nspace, int overspill)

{ static int len,i;

  if (nspace==COLSEP_RTFTAB) return fprintf(outfile," \\tab ");
  else if (nspace==COLSEP_TAB) 
    { if (fputc('\t',outfile)==EOF) return -1;
      else return 1;
    } 
  else if (nspace==COLSEP_COMMA) return fprintf(outfile,",");
  else if (nspace==COLSEP_LINE) return fprintf(outfile," | ");
  else if (nspace<0) return 0;
  else
    { if (overspill>0 && nspace>1)
        { len=nspace-overspill;
          if (len<1) len=1;
        }
      else len=nspace;
      for (i=0;i<len;i++)
        if (fputc(' ',outfile)==EOF) return -1;
      return len;
    }
}


int getbindata (FILE *infile, int n, double x[])

{ static int count;

  count=fread(x,sizeof(double),n,infile);
  if (count>0 && count<n) return FMTERR;
  if (feof(infile)) return EOF;
  if (ferror(infile)) return FMTERR;
  return OK;
}


int putbindata (FILE *outfile, int n, double x[])

{ static int count;

  count=fwrite(x,sizeof(double),n,outfile);
  if (count<n) return OUTERR;
  return OK;
}


double sgetnum (char *s, int *error)

{ static double value;
  static int nchar;

  if (s==NULL || sscanf(s,"%lf%n",&value,&nchar)!=1 || nchar!=strlen(s))
    *error=1;
  else *error=0;

  if (*error) return NODATA;
  return value;
}


int sputnum (char *s, int nmax, double x, int fwidth, int sigdig)

{ static char sbuff[LINE_MAX+1];
  static int len;

  if (fwidth==DEFAULT) fwidth=FWIDTH_DEF;
  if (x==NODATA) len=sprintf(sbuff,"%-*s",fwidth,NODATA_STR);
  else
    { if (sigdig==DEFAULT || sigdig<1) sigdig=SIGDIG_DEF;
      if (sigdig>LINE_MAX-6 || fwidth>LINE_MAX) return -1;
      x=roundval(x,sigdig);
      len=sprintf(sbuff,"%*.*g",fwidth,sigdig,x);
    }
  if (len>nmax) return -1;

  if (s) strcpy(s,sbuff);
  return len;
}


FILE *outfopen (char *fnamereq, char *outnamebuff, char **fname, int *ttb)

{ static int len;

  if (strcmp(fnamereq,".")==0)
    { *fname="-";
      *ttb=0; 
      return stdout;
    }
  else if (strcmp(fnamereq,"-")==0)
    { *fname="-";
      return stdout;
    }
  else if (strcmp(fnamereq,"-.ttb")==0)
    { *fname="-";
      *ttb=1;
      return stdout;
    } 
  else if (*ttb) *fname=ttbname(fnamereq,outnamebuff);
  else
    { len=strlen(fnamereq);
      *ttb=(len>=4 && strcmp(fnamereq+len-4,".ttb")==0);
      *fname=fnamereq;
    }
  return fopen(*fname,"w");
}


char *ttbname (char *root, char *outnamebuff)

{ static char *s;
  static int len;

  len=strlen(root);
  if (strcmp(root+len-4,".ttb")==0 ||
      strcmp(root+len-4,".TTB")==0 ||
      len+4>FILENAME_MAX) return root;

  s=outnamebuff;
  strcpy(s,root);
  strcpy(s+len,".ttb");
  return outnamebuff;
}


void freelist (struct datarec **head)

{ static struct datarec *p,*q;

  p=*head;
  while (p)
    { q=p->next;
      free(p->x);
      free(p->fieldlist);
      free(p->field);
      free(p);
      p=q;
    }
  *head=NULL;
}


void exiterror (int status, char *progname, char *s , int gcount)

{ static char *p;

  if (s==NULL) s="NULL";
  fprintf(stderr,"Error running %s: ",progname);
  switch (status)
    { case BADARG:
        fprintf(stderr,"space(s) in argument: %s\n",s);
        exit(EXIT_FAILURE);
      case OPENINERR:
        fprintf(stderr,"cannot open input file '%s'\n",s);
        exit(EXIT_FAILURE);
      case OPENOUTERR:
        fprintf(stderr,"cannot open output file '%s'\n",s);
        exit(EXIT_FAILURE);
      case FILELIMIT:
        fprintf(stderr,"too many files - can't open '%s'\n",s);
        exit(EXIT_FAILURE);
      case TRUNC:
        fprintf(stderr,"data truncation reading file '%s'",s);
        if (gcount>0) fprintf(stderr,", line %d\n",gcount);
        else if (gcount<0) fprintf(stderr,", data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case FLDTRUNC:
        fprintf(stderr,"too many fields in file '%s'",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
	exit(EXIT_FAILURE);
      case FNAMETRUNC:
	fprintf(stderr,"truncation extending output filename '%s'\n",s);
	exit(EXIT_FAILURE);
      case FMTERR:
        fprintf(stderr,"format error in file '%s'",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case BADDATA:
        fprintf(stderr,"unexpected data in file '%s'",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case VFMTERR:
        fprintf(stderr,"invalid variable name in file '%s'\n",s);
        exit(EXIT_FAILURE);
      case BADEXPR:
        fprintf(stderr,"invalid argument or expression: %s\n",s);
        exit(EXIT_FAILURE);
      case BADNUM:
        fprintf(stderr,"invalid number: %s\n",s);
        exit(EXIT_FAILURE);
      case BADFUNC:
        fprintf(stderr,"unknown function: %s\n",s);
        exit(EXIT_FAILURE);
      case NOVAR:
        fprintf(stderr,"no variables specified\n");
        exit(EXIT_FAILURE);
      case VARERR:
        p=strpbrk(s,"\\");
        if (p)
          { *(p++)='\0';
            fprintf(stderr,"variable '%s' not found in file '%s'\n",s,p);
          }
        else fprintf(stderr,"variable '%s' not found\n",s);
        exit(EXIT_FAILURE);
      case NOOUTVAR:
        fprintf(stderr,"no output variables\n");
        exit(EXIT_FAILURE);
      case OUTVARERR:
        fprintf(stderr,"too many output variables\n");
        exit(EXIT_FAILURE);
      case UNSORTED:
        fprintf(stderr,"file '%s' not sorted",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case DUPLICATE:
        fprintf(stderr,"duplicate data in file '%s'",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case OUTERR:
        fprintf(stderr,"writing to file '%s'\n",s);
        exit(EXIT_FAILURE);
      case OUTTRUNC:
        fprintf(stderr,"output data too long for file '%s'",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case RECLIMIT:
        fprintf(stderr,
          "too many records to process from file '%s'",s);
        if (gcount>0) fprintf(stderr," at line %d\n",gcount);
        else if (gcount<0) fprintf(stderr," at data record %d\n",-gcount);
        else fprintf(stderr,"\n");
        exit(EXIT_FAILURE);
      case MEMERR:
        fprintf(stderr,"memory allocation failure\n");
        exit(EXIT_FAILURE);
      case SPECIAL:
        fprintf(stderr,"%s\n",s);
        exit(EXIT_FAILURE);
      case EOF:
        fprintf(stderr,"unexpected end of file reading '%s'\n",s);
        exit(EXIT_FAILURE);
      case OK_INLINE:
        fprintf(stderr,
          "transposed table format used in file '%s' is not supported\n",s);
        exit(EXIT_FAILURE);
      default:
        fprintf(stderr,"status=%d\n",status);
        exit(EXIT_FAILURE);
    }
}


